/* 
 * MobileApplication.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 *  $Id: MobileApplication.Java,v 1.4 2000/02/08 23:42:55 tullmann Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */

/*
 * The Mobile test application for ANTSR(R)
 *
 * This is a simple application that uses a mobile protocol implementation to
 * forward packets from a home router to a foreign router, and finally to the
 * host.  The application takes arguments specifying the layout of the network
 * and which instance of the application should act as the server.  When the
 * mobile is on the home network, the router will just deliver packets directly
 * to it.  When the host is on the foreign network it will send out update
 * packets to store forwarding information on the foreign and home routers.
 * Any mobile capsules sent from a server will then use the forwarding info
 * on the routers to direct itself to the new location of the mobile.
 *
 *
 * Network Layout:
 *
 *                                       server
 *                                         |
 * mobile@foreign -- foreign router -- (internet) -- home router -- mobile@home
 *
 *
 * Output:
 *
 *  The output at the moment is pretty basic, it just prints out a message
 *  whenever it receives a capsule.  The thing to look for is that the 'from'
 *  address changes from the mobile@home address to the mobile@foreign address,
 *  and then changes back again, after the '*** Switching to' messages.
 */

package JANOS_ANTSR_MOBILE_APP_PKG;

import JANOS_NODEOS_PKG.AddressFormatException;

import JANOS_ANTSR_CORE_PKG.Application;
import JANOS_ANTSR_CORE_PKG.ByteArray;
import JANOS_ANTSR_CORE_PKG.Capsule;
import JANOS_ANTSR_CORE_PKG.Node;
import JANOS_ANTSR_CORE_PKG.NodeAddress;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;


/**
 * primitive Mobile test application
 *
 * @author David Wetherall
 */
public class MobileApplication
	extends Application
    implements Runnable
{
	int received;
	int iter;
	int mobile;
	int home;
	int foreign;
	int alternate;
	int udelay;
	int switchDelay;
	int switchCount;
	volatile boolean away;
  
	private final int LOG_LEVEL = 1;
	
	final public static String[] defaults = { "-update", "1000",
						  "-iter", "10000",
						  "-switchcount", "10000",
						  "-switch", "8" };
	
	/*
	 * Class that will send any update capsules when we're away
	 */
	PACKAGE class MAUpdate
		implements Runnable
	{
		MobileApplication mapp;

		private final int LOG_LEVEL = 1;

		/**
		 * Loop forever and send update capsules whenever we're away
		 */
		public void run()
		{
			int switcheroo = 1;
			int loop = 0;
		
			/* Loop for a while */
			while( iter >= 0 )
			{
				try
				{
				        /* Sleep for our update interval time */
					mapp.thisNode().sleep(mapp.udelay);
				}
				catch (InterruptedException e)
				{
					continue;
				}

				/* Switch from mobile to host every now and then... */
				if( (switchCount > 0) &&
				    (switcheroo % switchDelay == 0) )
				{
					switchCount--;
					mapp.away = !mapp.away;
					mapp.thisNode().log(LOG_LEVEL,
							    "*** Switching to " +
							    (mapp.away ?
							     "away" :
							     "home"));
				}
				switcheroo++;
				if (!mapp.away) continue;

				/*
				 * The mobile host is on the foreign network, send
				 * out an update packet
				 */
				mapp.send(new MobileUpdateCapsule(mapp.home,
								  mapp.foreign,
								  mapp.alternate));
				/* Track the number of times we've sent an update */
				loop++;
				mapp.thisNode().log(LOG_LEVEL,
						    "*** " + loop + " updates");
			}
		}

		/**
		 * Create an MAUpdate 
		 */
		PACKAGE MAUpdate(MobileApplication mapp)
		{
			this.mapp = mapp;
		}
	}

	synchronized public void receive(Capsule cap) {
		super.receive(cap);
		
		if (cap instanceof MobileCapsule)
		{
			/*
			 * We received a packet!  Dump out some information
			 * about it.
			 */
			received++;
			iter--;
			if( iter <= 0 )
				System.exit(0);
			MobileCapsule mcap = (MobileCapsule)cap;
			thisNode().log(LOG_LEVEL,
				       "Capsule #" + received
				       + " ("
				       + mcap.data.getByte(0)
				       + ","
				       + (char)mcap.data.getByte(1)
				       + ") from "
				       /*
					* The next field will be set to the
					* destination address, which will
					* either be the mobile host, or the
					* foreign network address for the
					* mobile.
					*/
				       + NodeAddress.toString(mcap.next)
				       );
		}
		else
		{
			thisNode().log(LOG_LEVEL,
				       "Received non mobile capsule "
				       + cap);
		}
	}

	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);
		try
		{
			for (int i = 0; i < k.length(); i++) {
				if (k.key(i).equals("-mobile")) {
					mobile = NodeAddress.
						fromString(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-home")) {
					home = NodeAddress.
						fromString(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-foreign")) {
					foreign = NodeAddress.
						fromString(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-alternate")) {
					alternate = NodeAddress.
						fromString(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-update")) {
					udelay = Integer.parseInt(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-iter")) {
					iter = Integer.parseInt(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-switch")) {
					switchDelay =
						Integer.parseInt(k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-switchcount")) {
					switchCount =
						Integer.parseInt(k.arg(i));
					k.strike(i);
				}
			}
		}
		catch( AddressFormatException afe )
		{
			throw new InvalidKeyArgsException("HACK HACK HACK: "
							  + afe.getMessage());
		}

		super.setArgs(k);
	}

	/**
	 * Send packets to the mobile host, hopefully they get routed correctly
	 */
	public void run()
	{
		ByteArray ba;
		byte count = 0;

		/* Construct some data to send in a packet */
		ba = new ByteArray(2);
		ba.setByte(1, (byte)'a');
		while( iter != 0 )
		{
			iter--;
			/* Add something that will change for each packet */
			ba.setByte(0, count);
			try
			{
				/* Sleep for a bit and then send a packet */
				thisNode().sleep( 50 );
				this.send( new MobileCapsule(mobile,
							     (short)getPort(),
							     (short)getPort(),
							     ba) );
			}
			catch( InterruptedException ie ) 
			{
				;
			}
			count++;
		}
	}
	
	public void start() 
		throws Exception 
	{
		thisNode().register(new MobileProtocol());
		
		if (mobile == thisNode().getAddress())
		{
			/*
			 * We're the mobile host, startup the update thread
			 * that will act as the client and change from home to
			 * away, and send out MobileUpdateCapsules
			 */
			thisNode().threadStart(new MAUpdate(this));
		} else {
			/*
			 * We're the source host, just start a thread for this
			 * object that constantly sends objects off to the
			 * mobile host.
			 */
			thisNode().threadStart(this);
		}
	}
}
