/* 
 * MulticastApplication.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package JANOS_ANTSR_MULTICAST_APP_PKG;

import JANOS_ANTSR_CORE_PKG.*;
import JANOS_ANTSR_UTIL_PKG.*;

/**
 * Thread to send out periodic subscribe requests for the multicast group.
 */
class MASubscribe
	implements Runnable 
{
	MulticastApplication mapp;
	
	public void run() {
		while (true) {
			try {
				Thread.sleep(mapp.sdelay);
			} catch (InterruptedException e) {
				continue;
			}
			
			int target = mapp.target;
			System.out.println("Sending subscribe capsule.");
			mapp.send(new MulticastSubscribeCapsule(
				mapp.target, mapp.group,
				(short)mapp.getPort()));
		}
	}
	
	MASubscribe(MulticastApplication mapp) {
		this.mapp = mapp;
	}
}

/**
 * Thread to send out periodic data capsules to the multicast group.
 */
class MASend
	implements Runnable
{
	MulticastApplication mapp;
	int count = 0;
	
	public void run() {
		while (true) {
			try {
				Thread.sleep(mapp.sdelay);
			} catch (InterruptedException e) {
				continue;
			}
			
			/* Increment multicast count. */
			count++;
			
			int target = mapp.target;
			System.out.println("Sending multicast capsule # "
					   + count);
			byte[] data = { 0x01, 0x02, 0x03, 0x04,
					(byte) (count >> 24),
					(byte) (count >> 16),
					(byte) (count >> 8),
					(byte) (count) };
			mapp.send(new MulticastCapsule(mapp.group, mapp.target,
						       (short)mapp.getPort(),
						       new ByteArray(data)));
		}
	}
	
	MASend(MulticastApplication mapp) {
		this.mapp = mapp;
	}
}

/**
 * The Multicast test application for ANTSR(R).
 *
 * Input:
 *   -delay n	Number of milliseconds to wait between multicast subscriptions
 *		or number of milliseconds to wait between multicasts.
 *   -target a	Address of multicast sender (target for subscriptions).
 *   -group a	Multicast group "address".
 *
 * Output:
 *   This program generates output indicating when capsules are received.
 *
 * @author David Wetherall
 */
public class MulticastApplication
	extends Application
{
	int target;
	int group;
	int sdelay;
	int scount;
	
	final private static String[] DEFAULTS = { "-delay", "1000" };
	
	synchronized public void receive(Capsule cap) {
		super.receive(cap);
		
		if (cap instanceof MulticastCapsule) {
			MulticastCapsule mc = (MulticastCapsule)cap;
			
			if (mc.data.length() < 8)
				System.out.println("Received bad multicast capsule!");
			else {
				int count = ((int)mc.data.getByte(4)) << 24
					    | ((int)mc.data.getByte(5)) << 16
					    | ((int)mc.data.getByte(6)) << 8
					    | (mc.data.getByte(7) & 0xff);
				System.out.println("Received multicast capsule #" + count);
			}
		} else if (cap instanceof MulticastSubscribeCapsule) {
			scount++;
			System.out.println("Received multicast subscription #" + scount);
		}
	}
	
	
	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(DEFAULTS);
		
		try {
			for (int i = 0; i < k.length(); i++) {
				if (k.key(i).equals("-target")) {
					target = NodeAddress.fromString(
						k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-group")) {
					group = NodeAddress.fromString(
						k.arg(i));
					k.strike(i);
				} else if (k.key(i).equals("-delay")) {
					sdelay = Integer.parseInt(k.arg(i));
					k.strike(i);
				}
			}
		}
		catch (Exception e) {
			throw new InvalidKeyArgsException("HACK HACK HACK: "
							  + e.getMessage());
		}
		
		super.setArgs(k);
	}
	
	public void start() 
		throws Exception 
	{
		/* Register the multicast protocol. */
		thisNode().register(new MulticastProtocol());
		
		if (target == thisNode().getAddress()) {
			System.out.println("Starting source thread...");
			thisNode().threadStart(new MASend(this));
		} else {
			System.out.println("Starting client thread...");
			thisNode().threadStart(new MASubscribe(this));
		}
	}
	
	public MulticastApplication() {
		super();
	}
}

