/*
 * AppBooster.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.ThreadPool;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.CommSpaceElement;

import JANOS_IO_PKG.Log;

import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Cannot be made public.
 * XXX: well, at least it can't be instantiated by the unwashed masses.
 *
 * CommSpaceHandle should point to a valid CSH for
 * the Booster class, plus:
 *   [2] this will write the App port to this offset.
 */
public class AppBooster
	extends Booster
{
	private ThreadPool threadPool;
	
	/**
	 * Must have the appropriate key to create an AppBooster.
	 * Only the core can actually create Keys, so only trusted
	 * code can create AppBoosters...
	 *
	 * @exception SecurityException if the key is inappropriate
	 * @exception NullPointerException if the key is null
	 */
	public AppBooster(JANOS_NODEOS_PKG.ClassInstantiationKey key)
		throws SecurityException
	{
		key.checkEquals(AppBooster.class);
	}

	protected void run(final Flow flow,
			   ThreadPool threadPool,
			   String[] args,
			   CommSpaceHandle commSpaceHandle)
		throws Exception
	{
		DEBUGONLY(flow.setName("Flow for App " +args[0]));

		this.threadPool = threadPool;
		// XXX ClassNotFoundException, ClassCastException, etc...
		final Class appClass = flow.loadClass(args[0]);
		final Application app = (Application)(appClass.newInstance()); 
		app.setName(args[0]);
		app.setArgs(args, 1);

		// setupLocalNode was inherited from Booster
		final Node localNode = setupLocalNode(flow, commSpaceHandle);

		app.setNode(localNode);

		// Install the app in the global application table
		localNode.attachApplication(app);

		TRACE(Trace.ANTSR_APP, "App " +args[0]+ " attached at port " +app.getPort());

		threadPool.threadStart(new Runnable()
			{
				public void run()
				{
					processPacket(flow,
						      app,
						      localNode);
				}
			}
		).setName("Packet Processor for " +args[0]);
		
		// Fire off the app's entry point.
		app.start();
	}

	/**
	 * Delivers packets from a protocol flow to an application.
	 */
	private void processPacket(Flow flow, Application app, Node node)
	{
		CommQueue queue = app.getPacketQueue();

		/*
		 * Loop processing packets from protocol flows.
		 * <p>
		 * XXX Looping on whether the threadpool was destroyed or not
		 * doesn't really work since we'll be stuck elsewhere and
		 * won't notice unless we get a packet...  Oh well, this
		 * should be taken care of by other stuff.
		 */
		while(!threadPool.destroyed())
		{
			/* Grab the next capsule from our CommQueue */
			Capsule cap = Capsule.dequeueCapsule(node, queue);

			/* Pass to the app for processing */
			app.receive(cap);
		}
	}
}
