/*
 * CapsuleChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.ThreadPool;

import JANOS_IO_PKG.Log;

import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * The handler for incoming ANTSR packets.  They get decoded into
 * Capsules and evaluted here.
 */
PACKAGE final class CapsuleChannel
	implements DeliverFunc
{
	PACKAGE Class cl;
	PACKAGE InChannel inChannel;
	private ANTSRDemultiplexKey classifier;

	PACKAGE CapsuleChannel(CapsuleClass cc)
	{
		this(cc, Flow.currentFlow().threadPool());
	}

	/**
	 * Construct a receive channel for cc Capsules.  Implicitly
	 * constructs an ANTSRDemultiplexKey to match the right packets.
	 */
	PACKAGE CapsuleChannel(CapsuleClass cc, ThreadPool threadPool)
	{
		this(cc.cl, new ANTSRDemultiplexKey(cc.methodID), threadPool);
	}

	PACKAGE CapsuleChannel(Class cl, ANTSRDemultiplexKey classifier)
	{
		this(cl, classifier, Flow.currentFlow().threadPool());
	}

	PACKAGE CapsuleChannel(Class cl, ANTSRDemultiplexKey classifier, ThreadPool threadPool)
	{
		Flow flow = Flow.currentFlow();
		this.cl = cl;
		this.classifier = classifier;
		TRACE(Trace.ANTSR_CAPSULECHANNEL, "new capsule channel: class=" +cl+ "; classifier=" +classifier);
		inChannel = new InChannel(
			flow,
			classifier,
			threadPool,
			this);
	}

	/**
	 * Handler for packets delivered to this Channel.
	 * <p>
	 * Receives a BufferHandle and decodes the packet
	 * into a Capsule.
	 * <p>
	 * Assumes packet[0] is the first byte of the ANEP
	 * header...
	 * <p>
	 * XXX this is atrocious code.	Rewrite it now.
	 */
	public void receive(BufferHandle packet, int payloadStart)
		throws IOException
	{
		assert(packet);
		assert(this.cl);
		assert(this.classifier);

		// XXX
		//
		// XXX This is old code from before I knew what was
		// going on.  Don't trust it.  :) OTOH, it could be
		// right...
		// 
		// XXX


		// XXX Xdr should wrap a BufferHandle better than this!
		byte[] buf = new byte[packet.getLength()];
		packet.getBytes(0, buf, 0, buf.length);
		Xdr xdr = new Xdr(buf);
		TRACE(Trace.ANTSR_CAPSULECHANNEL, "buf.length = " +buf.length+ ", payloadStart = " +payloadStart);

		Capsule c;
		try
		{
			TRACE(Trace.ANTSR_CAPSULECHANNEL, "capsule channel instantiating " +cl);
			c = (Capsule)cl.newInstance();
			// c.mid = classifier.getMID();
		}
		catch (Exception e)
		{
			panic("blew up creating a new instance of " +cl.getName());
		}

		// increment xdr through ANEP crap
		// also skip the ANTSR typeid

		// if this is a special packet find the actual payload start.
		// "special packets" went through UnknownMID processing.
		if(payloadStart == UnknownMidPackets.NEVER_DEMULTIPLEXED)
		{
			payloadStart = classifier.matches(buf, 0, buf.length);
			assert(payloadStart > 0);
		}

		xdr.skip(classifier.whereDoesTheCapsuleDecodingActuallyStart(
			payloadStart));

		// Decode the xdr into the Capsule
		c.decodeWith(xdr); // XXX exceptions??

		// Evaluate the new Capsule
		Node node = Node.currentNode();
		try
		{
			boolean sentIt;
			sentIt = c.evaluate(node);
		}
		catch (Throwable th)
		{
			// XXX use Trace/Log/etc...
			node.log(Log.ERR, "Capsule " +c+ " exploded: " +th.getMessage());
			node.logStackTrace(Log.ERR, th);
		}
	}

	public String toString()
	{
		return "CapsuleChannel[cl="
			+ this.cl.getName()
			+ "; inChannel= XX"
			// + this.inChannel // XXX circular toString() references
			+ "; classifier="
			+ this.classifier
			+ "]";
	}
}
