/* 
 * MIDIDResponseCapsule.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * MID ID Response Capule.
 * <p>
 * MIDIDResponseCapsules are only created by a MIDIDRequestCapsule
 * that has been delivered to a node that can satisfy the request.
 * MIDIDResponseCapsules contain information describing the MID.
 * <p>
 * A MIDIDResponseCapsule packet contains the following fields (on top of the
 * base Capsule fields):
 * <ul>
 *   <li> requestMID: MID that initiated the request
 *   <li> pid: PID associated with the MID.
 *   <li> userInfo: userInfo (see UserID).
 * </li>
 * 
 * @see DLProtocol
 * @see MIDIDRequestCapsule
 * @see CodeGroup
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE class MIDIDResponseCapsule
	extends Capsule 
{
	// WARNING: this is a builtin capsule, whose version number
	// must be manually kept in synch with  its functionality
	final public static String VERSION = "JANOS MIDIDResponseCapsule July 17, 1999";

	final private static byte[] MID = findMID(JANOS_ANTSR_CORE_PKG_S + ".MIDIDResponseCapsule");
	final private static byte[] PID = findPID(JANOS_ANTSR_CORE_PKG_S + ".MIDIDResponseCapsule");

	protected byte[] mid()
	{
		return MID;
	}

	protected byte[] pid()
	{
		return PID;
	}

	/**
	 * The MID that initiated the request.
	 */
	private TypeID requestMID;

	/**
	 * PID associated with the requestMID.
	 */
	private TypeID pid;

	/**
	 * GID associated with the requestMID.
	 */
	private TypeID gid;

	/**
	 * User information associated with the PID.
	 */
	private UserID userInfo;

	public int length() 
	{
		return super.length() +
			Xdr.TYPEID +		// groupID
			Xdr.TYPEID +		// payloadID
			Xdr.TYPEID +		// protocolID
			userInfo.length();
	}
  
	public Xdr encode() 
	{
		Xdr xdr = super.encode();
		
		xdr.PUT(gid);
		xdr.PUT(requestMID);
		xdr.PUT(pid);
		userInfo.xdrPut(xdr);

		return xdr;
	}

	public Xdr decode()
	{
		Xdr xdr;

		xdr = super.decode();

		gid = xdr.TYPEID();
		requestMID = xdr.TYPEID();
		pid = xdr.TYPEID();
		this.userInfo = UserID.xdrGet(xdr);
    
		return xdr;
	}
  
	/**
	 * Evalute a MIDIDResponseCapsule at this node.
	 * <p>
	 * XXX should deal with unanticipated MIDIDResponseCapsules better.
	 */
	public boolean evaluate(Node node)
	{
		boolean retval;
		
		TRACE(Trace.ANTSR_DYNLOAD, "received MIDIDResponseCapsule " +requestMID);

		/*
		 * Don't bother to double check that I asked for this
		 * MIDIDResponseCapsule.  If something is trying to waste
		 * our time, they can just do a better (and undetectable)
		 * job by sending fake packets with unknown MIDs.
		 */

		// XXX Double check that the MIDIDResponse is valid...
		// (i.e., the UserInfo...)

		if( node.getAddress() == getDst() )
		{
			/* Remove the pending request for this mid */
			node.getCache().remove(requestMID);
			PrimordialNode.addMID(pid, userInfo, requestMID, gid, getSrc());
			retval = true;
		}
		else
		{
			retval = node.routeForNode(this, getDst());
		}
		return retval;
	}
  

	/**
	 * Construct a blank MIDIDResponseCapsule.
	 * <p>
	 * This is only invoked when a MIDIDResponseCapsule capsule is
	 * received at a node and the node must construct a capsule
	 * before decoding the packet into it.
	 */
	PACKAGE MIDIDResponseCapsule()
	{
	}
	
	/**
	 * Construct a MIDIDResponseCapsule to send back in response
	 * to a MIDIDRequestCapsule.  Only invoked by
	 * MIDIDRequestCapsule.evaluate().
	 *
	 * @see MIDIDRequestCapsule#evaluate(ants.Node)
	 */
	PACKAGE MIDIDResponseCapsule(int dst,
				     TypeID unknownMID,
				     TypeID pid,
				     TypeID gid,
				     UserID userInfo)
	{
		setDst(dst);
		this.requestMID = unknownMID;
		this.pid = pid;
		this.gid = gid;
		this.userInfo = userInfo;
	}
	
	public String toString()
	{
		StringBuffer buf = new StringBuffer(100);

		buf.append(super.toString());
		buf.append("+MIDIDResponse[requestMID=");
		buf.append(requestMID);
		buf.append("; pid=");
		buf.append(pid);
		buf.append("; gid=");
		buf.append(gid);
		buf.append("]");
		return buf.toString();
	}
}
