/* 
 *  $Id: PrimordialNode.Java,v 1.3 2000/01/26 23:30:26 tullmann Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;

import JANOS_ROUTE_PKG.RouteEntry;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.Credentials;
import JANOS_NODEOS_PKG.ClassSpec;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.FlowStartSpec;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.ThreadPool;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JANOS_NODEOS_PKG.ProtocolRegister;

import JANOS_IO_PKG.Log;
import JANOS_IO_PKG.LogListener;
import JANOS_IO_PKG.DatagramLogListener;

import java.io.IOException;

import java.net.InetAddress; // XXX used for hostname in control console printout

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * The root Node object that boots the system and tracks system-wide
 * state.  This is as opposed to the per-flow Node interface
 * that every protocol/application sees.
 * <p>
 * XXX move the ConfigurationManager stuff into here.
 * <p>
 * XXX Move the command line option stuff OUT of here.  Make start()
 * method take functions, or make them static method (e.g, if can be
 * changed at run-time).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE class PrimordialNode
{
	/**
	 * Arguments the PrimordialNode understands and their defaults.
	 */
	private final static String[] defaults = {"-consoleport",  null,
						  "-msgprefix",    null,
						  "-onetrick",     null,
						  "-log",          "0",
						  "-routes",       null,
						  "-queryport",    null};

	/**
	 * Port to start a console thread on this port (if > -1).
	 */
	private static int consolePort  = -1;
	private static int watchDogPort = -1;
	private static int queryPort    = -1;

	private static long startupTime = System.currentTimeMillis();

	/**
	 * Is this a one-trick node?
	 *
	 * @see #oneTrick()
	 */
	private static boolean oneTrickPony = false;

	private final static Log logger = new Log(null);

	private static class StdoutListener
		implements LogListener
	{
		public void logMsg(int level, String msg)
		{
			System.out.println(msg);
		}
	}
	
	static
	{
		logger.addListener(new StdoutListener());
	}
	
	private static String msgPrefix = null;

	private static String routefile = "ping.routes"; // XXX XXX XXX
	private static boolean updateRoutes = false;
	private static RouteTable routeTable;
	private static CommSpaceHandle neighborsSpace;

	/**
	 * 32-bit logical address of this Node.
	 * <p>
	 * XXX replace with a real Address object?
	 */
	private static int address;

	/**
	 * Physical address as an "opaque" String.  ("only" addr can
	 * parse it...
	 */
	private static String physAddr = null;

	// Maps process IDs to CommQueues:
	private static final Hashtable pidToFlowQueue = new Hashtable();

	/**
	 * Global cache of packets for MIDs that haven't been loaded/resolved.
	 * <p>
	 * The Dead Letter Office.
	 */
	PACKAGE static final UnknownMidPackets unknownMidPackets =
		new UnknownMidPackets();


	/**
	 * No instance of this class is ever created.
	 */
	private PrimordialNode()
	{
	}

	PACKAGE static CommQueue getProtocolQueue(TypeID pid)
	{
		return (CommQueue)pidToFlowQueue.get(pid);
	}

	PACKAGE static long getStartupTime()
	{
		return startupTime;
	}
	
	private static RouteTable loadRouteTableImpl(int addr, 
						     boolean updateRoutes, 
						     String routeTableFile)
	{
		RouteTable rti = null;
		String rtClassname;
		Class rtiClass;

		rtClassname = JANOS_ANTSR_ROUTE_PKG_S 
			+(updateRoutes?".DynamicRouteTable":".StaticRouteTable");
		try
		{
			rtiClass = Class.forName(rtClassname);
			rti = (RouteTable)rtiClass.newInstance();
			rti.configure(addr, updateRoutes, routeTableFile);
		}
		catch (NoSuchNodeAddress nsne)
		{
			FATALEXCEPTION(nsne, "Bad address in " +routeTableFile);
		}
		catch (AddressFormatException afe)
		{
			FATALEXCEPTION(afe, "Bad route configuration file " +routeTableFile);
		}
		catch (IOException ioe)
		{
			FATALEXCEPTION(ioe, "Blew up reading " + routeTableFile);
		}
		catch (InstantiationException ie)
		{
			FATALEXCEPTION(ie, rtClassname+ " cannot be instantiated.");
		}
		catch (IllegalAccessException iae)
		{
			FATALEXCEPTION(iae, rtClassname+ " cannot be instantiated.");
		}
		catch (ClassCastException cce)
		{
			FATALEXCEPTION(cce, rtClassname+ " is probably does not implement RouteTable.");
		}
		catch (ClassNotFoundException cnfe)
		{
			FATALEXCEPTION(cnfe, "While loading route table class: " +rtClassname);
		}

		return rti;
	}


	/**
	 * Start the node on its way.
	 */
	PACKAGE static void start(Vector appConfigs, Vector extConfigs)
		throws IOException, AddressFormatException, NoSuchNodeAddress,
		InterruptedException
	{
		assert(appConfigs);
		assert(extConfigs);

		/*
		 * Define the local packages for ANTSR.  These are
		 * packages that will be loaded into the root (e.g,
		 * system) classloader.  All others will be loaded on
		 * a per-flow basis.
		 *
		 * Look at the notes in core.FlowClassLoader.
		 */
		Flow.currentFlow().addSystemPackage(JANOS_ANTSR_CORE_PKG_S);
		Flow.currentFlow().addSystemPackage(JANOS_ANTSR_UTIL_PKG_S);

		/*
		 * XXX Route packge is a "system" package so that the RouteTable
		 * class gets loaded by the System class loader and is thus shareable
		 * between flows...  not ideal.
		 */
		Flow.currentFlow().addSystemPackage(JANOS_ANTSR_ROUTE_PKG_S);

		/*
		 * Setup the route table.  Initialize from the routefile
		 * argument.
		 */
		routeTable = loadRouteTableImpl(getAddress(), updateRoutes, routefile);
		assert(routeTable);

		routefile = null; // drop the route file name

		/*
		 * Get the list of neighbors into a form the child app
		 * can use.
		 */
		neighborsSpace = fillNeighborSpace();
		
		/*
		 * Create a Node interface for the builtin protocols.
		 */
		Node node = new Node(
			Flow.currentFlow(),
			neighborsSpace,
			address,
			unknownMidPackets);

		assert(address != 0);

		/*
		 * XXX Perhaps the protocols started in here should be created
		 * inside their own flows?  Or do they need access to some
		 * global state?
		 */
		try
		{
			/*
			 * Create the DLProtocol.  
			 *
			 * Manually associate it with a key that matches
			 * just the ANTSR header.  This way it will get all
			 * packets that aren't matched by more specific
			 * filters.
			 *
			 * XXX runs with the same ThreadPool as other root flow tasks...
			 * ... might be okay actually (nothing else for those threads).
			 */
			DLProtocol dlp = new DLProtocol();
			
			TRACE(Trace.ANTSR_STARTUP, "Installing " +dlp);

			/*
			 * Activate the MIDIDRequest/Response parts of
			 * the DLProtocol.  (The other parts
			 * aren't activated because they're inert --
			 * see BuiltinProtocol.)
			 */
			node.register(dlp);

			/*
			 * Setup the DataProtocol.  Self-forwarding data
			 * packet service is provided by the Node.
			 *
			 * XXX this should be tuned & optimized.
			 */
			DataProtocol dp = new DataProtocol();
			TRACE(Trace.ANTSR_STARTUP, "Installing " +dp);
			node.register(dp);

			// XXX unknown ANEP packet forwarder--- a cut-through channel!.
		}
		catch (Exception e)
		{
			FATALEXCEPTION(e, "Builtin protocol setup failure.");
		}
		
		/*
		 * Start the dynamic routing (if necessary).
		 */
		if (updateRoutes)
		{
			// Start the app that maintains the routing table
			AppConfig dynrouteApp = new AppConfig(JANOS_ANTSR_ROUTE_PKG_S + ".DynamicRouteApp",
							      null, 0);
			fireUpNewAppFlow(dynrouteApp);
		}

		/*
		 * Start up the applications on this node.
		 */
		Enumeration apps = appConfigs.elements();
		while(apps.hasMoreElements())
		{
			AppConfig a = (AppConfig)apps.nextElement();
			fireUpNewAppFlow(a);
		}

		
		/*
		 * XXX MIT ANTS supports Node extensions and installs
		 * them at this point.  Since they're unused, we ignore
		 * them for now.
		 */
		assert(extConfigs.size() == 0);


		/*
		 * Start the console.
		 */
		startConsole(node, consolePort);
		startWatchDog(node, watchDogPort);
		if( queryPort > -1 )
		{
			DatagramLogListener dll =
				new DatagramLogListener();

			try
			{
				dll.start(queryPort);
				logger.addListener(dll);
			}
			catch(Exception e)
			{
			}
		}
	}
	

	private static void startConsole(Node node, int consolePort)
	{
		if (consolePort > -1)
		{
			Console con = null;
			int tries = 1;

			while ((tries != 0) && (tries < 15))
			{
				try
				{
					con = new Console(node, consolePort);
					tries = 0;
				}
				catch (IOException ioe)
				{
					tries++;
					consolePort++;
				}
			}

			if (con != null)
			{
				Flow.currentFlow().threadPool().
					threadStart(con).
					setName("Root Console");
				try
				{
					String hostname;
					hostname = java.net.InetAddress.getLocalHost().getHostName(); // XXX
					node.log(Log.NOTICE,
						 "Started a control console on "
						 +hostname+ ":" +con.getPort());
				}
				catch (java.net.UnknownHostException uhe)
				{
					node.log(Log.NOTICE,
						 "Started a control console on localhost:" +
						 con.getPort());
				}
			}
			else
			{
				node.log(Log.NOTICE,
					 "Unable to start a control console on port " +
					 con.getPort());
			}
		}
	}

	private static void startWatchDog(Node node, int watchDogPort)
	{
		if(watchDogPort > -1)
		{
			WatchDog wd = null;
			int tries = 1;

			while( (tries != 0) && (tries < 15) )
			{
				try
				{
					wd = new WatchDog(node, watchDogPort);
					tries = 0;
				}
				catch(IOException e)
				{
					tries++;
				}
			}
			if( wd != null )
			{
				Flow.currentFlow().threadPool().
					threadStart(wd).
					setName("Root Watchdog");
				try
				{
					String hostname;
					
					hostname = java.net.InetAddress.getLocalHost().getHostName(); // XXX
					node.log(Log.NOTICE,
						 "Started a watch dog on port "
						 +hostname+ ":" +wd.getPort());
				}
				catch(java.net.UnknownHostException e)
				{
					node.log(Log.NOTICE,
						 "Started a watch dog on localhost:" +
						 wd.getPort());
				}
			}
			else
			{
				node.log(Log.NOTICE,
					 "Unable to start a watch dog on port "
					 + watchDogPort);
			}
		}
	}
	
	/**
	 * Setup one outchan direct to each neighbor node.
	 * Enter each into a hashtable indexed by the logical
	 * address of the neighbor.
	 * <p>
	 * XXX record the logical address of each neighbor in the
	 * neigborsSpace commSpace.  Child protocols and apps
	 * will create a set of OutChannels from this info.
	 * Seems rather weak and ugly, though.  Perhaps push-able
	 * OutChannelHandles would be a good idea?
	 * <p>
	 * XXX this is now redundant with the per-node outchannels,
	 * no?
	 * <p>
	 * XXX Its really, really sad that I call this once for each
	 * Flow I create.  Kinda stupid, no?
	 */
	PACKAGE static CommSpaceHandle fillNeighborSpace()
		throws IOException
	{
		Enumeration neighbors = routeTable.getNeighbors();

		TRACE(Trace.ANTSR_ROUTEINFO|Trace.ANTSR_STARTUP, "I have " +routeTable.neighborCount()+ " neighbor(s).");

		CommSpaceHandle space = CommSpaceHandle.newCommSpace(
			routeTable.neighborCount() * 2 + 1);

		int idx = 0;
		while (neighbors.hasMoreElements())
		{
			RouteEntry r = (RouteEntry)neighbors.nextElement();

			// Neighbor table should only have immediate neighbors
			assert(r.getLogicalDest() == r.getLogicalNext());
			
			TRACE(Trace.ANTSR_ROUTEINFO, "Have immediate neighbor " +r);

			space.write(idx++,
				new CommSpaceElement(r.getLogicalDest()));
			space.write(idx++,
				new CommSpaceElement(r.getPhysicalNext()));
		}
		space.write(idx++, new CommSpaceElement(0));
		
		// XXX ... its bad if this node has no immediate neighbors, m'Kay.
		if (idx == 1)
		{
			message("ALERT: This node has no immediate, static neighbors!");
		}

		return space;
	}

	PACKAGE static RouteTable getRouteTable()
	{
		return routeTable;		// XXX access controls?
	}

	/**
	 * Return the logical ANTSR address of the next node
	 * on the path to the given addr.
	 */
	PACKAGE static int nextHopToNode(int addr)
	{
		// assert(addr != ..);
		RouteEntry re = routeTable.getRoute(addr);
		if (re == null)
		{
			TRACE(Trace.ANTSR_ROUTEINFO, "Couldn't get route for " +Address.toString(addr));
			return 0;
		}

		return re.getLogicalNext();
	}


	/*
	 * Parse a set of KeyArgs arguments.  Adds '-routes
	 * <filename>', '-log <int>' and '-updateRoutes'
	 * arguments.  Will throw an InvalidKeyArgsException if any
	 * arguments are not understood by Node or its superclases.
	 * <p>
	 * Defaults '-updateRoutes' to false.  Defaults route
	 * filename to null.
	 *
	 * @see #defaults
	 */
	PACKAGE static void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);
		
		for (int i = 0; i < k.length(); i++) 
		{
			if (k.key(i).equals("-routes")) 
			{
				routefile = k.arg(i);
				k.strike(i);
			}
			else if (k.key(i).equals("-msgprefix"))
			{
				msgPrefix = k.arg(i);
				logger.setPrefix(msgPrefix);
				k.strike(i);
			}
			else if (k.key(i).equals("-onetrick"))
			{
				oneTrickPony = true;
				k.strike(i);
			}
			else if (k.key(i).equals("-consoleport"))
			{
				// TRACE(Trace.ANTSR_STARTUP, "Parsing consoleport: " +k.arg(i)+ ".");
				if (k.arg(i) != null)
					consolePort = Integer.parseInt(k.arg(i));
				k.strike(i);
			}
			else if (k.key(i).equals("-updateRoutes"))
			{
				// updateRoutes = Boolean.valueOf(k.arg(i)).booleanValue();
				updateRoutes = true;
				k.strike(i);
			}
			else if (k.key(i).equals("-log")) 
			{
				// XXX ignore -log for now
				k.strike(i);
			}
			else if (k.key(i).equals("-watchdog"))
			{
				if (k.arg(i) != null)
					watchDogPort =
						Integer.parseInt(k.arg(i));
				k.strike(i);
			}
			else if (k.key(i).equals("-queryport"))
			{
			       if (k.arg(i) != null)
				       queryPort = Integer.parseInt(k.arg(i));
			       k.strike(i);
			}
		}

		k.noMore();
	}

	PACKAGE static boolean oneTrick()
	{
		return oneTrickPony;
	}

	/**
	 * Shut down the entire VM.  Only invoked (occasionally)
	 * by Node.shutdown().
	 */
	PACKAGE static void shutdown()
	{
		JANOS_NODEOS_PKG.Node.shutdown();
	}

	public static int getAddress()
	{
		return address; // XXX
	}

	/**
	 * Set this node's logical address.  Only invoked
	 * once during configuration of the system
	 * after the Node object is created.
	 *
	 * @see ConfigurationManager#parseCommands()
	 */
	PACKAGE static void setAddress(int address)
	{
		PrimordialNode.address = address;
	}	

	/**
	 * Set the physical address of this node.  Needed by the
	 * route table update algorithms.  Not parsed.  Only
	 * understood by the core.Address class.
	 *
	 * @param addr the String representation of this node's physical address.
	 */
	PACKAGE static void setPhysicalAddress(String addr)
	{
		assert(physAddr == null); // XXX only one addr per node!  Gak!
		physAddr = addr;
	}

	PACKAGE static String getPhysicalAddress()
	{
		assert(physAddr != null);
		return physAddr;
	}

	PACKAGE static void message(String msg)
	{
		logger.logMsg(0, msg);
	}

	PACKAGE static void message(int level, String msg)
	{
		logger.logMsg(level, msg);
	}

	PACKAGE static void dumpStackTrace(Throwable exc)
	{
		logger.logThrowable(exc);
	}

	/**
	 * Add the given MID to the set understood by this node.
	 * This creates a flow to handle the mid (using the
	 * resource defined by the userInfo field).
	 * <p>
	 * XXX Currently the whole method is synchronized.  This
	 * protects the race-prone lookup/create in the pidToFlowQueue
	 * map.  It doesn't need to protect the fireUpNewProtocolFlow.
	 *
	 * @param pid the PID of the new flow.
	 * @param userInfo the user identity/credentials of the new flow
	 * @param mid the MID that caused the flow to get loaded.
	 * @param prevNode the node that sent the initial MID-labeled packet.
	 */
	PACKAGE static synchronized void addMID(TypeID pid,
						UserID userInfo,
						TypeID mid,
						TypeID gid,
						int prevNode)
	{
		CommQueue queue = (CommQueue) pidToFlowQueue.get(pid);

		ProtocolRegister.register(
			pid, new ProtocolRegister.Owner(pid.toString(),
							"Downloaded"));
		if(queue == null)
		{
			TRACE(Trace.ANTSR_DYNLOAD, "addMID(pid=" + pid +") is new.");
			queue = new CommQueue();
			pidToFlowQueue.put(pid, queue);
			try
			{
				fireUpNewProtocolFlow(pid, queue);
			}
			catch (Throwable th)
			{
				FATALEXCEPTION(th, "Uh-oh.  fireUpNewProtocolFlow() blew up.");
			}
		}
		else
		{
			TRACE(Trace.ANTSR_DYNLOAD, "addMID(pid=" + pid +") Q already setup...");
		}

		/*
		 * Even if the pid is already registered, the MID
		 * probably isn't.  Push MID info into the queue for
		 * the new flow to snack on.
		 */
		queue.enqueue(new CommSpaceElement(ProtocolBooster.NEW_MID));
		queue.enqueue(new CommSpaceElement(mid.unwrap()));
		queue.enqueue(new CommSpaceElement(gid.unwrap()));
		queue.enqueue(new CommSpaceElement(prevNode));
	}

	/**
	 * Install a protocol, this will start up the a new flow if necessary
	 * and load in the code.
	 *
	 * @param pid The protocol ID to be loaded
	 * @param name The class name of the protocol to install
	 */
	PACKAGE static synchronized void installProtocol(TypeID pid,
							 String name)
	{
		CommQueue queue = (CommQueue)pidToFlowQueue.get(pid);

		/* Find the current owner of the protocol if there is one */
		ProtocolRegister.register(
			pid, new ProtocolRegister.Owner(name,
							"Local"));
		if( queue == null )
		{
			/*
			 * We're the owner now, make a new queue and start up a
			 * new flow
			 */
			queue = new CommQueue();
			pidToFlowQueue.put(pid, queue);
			try
			{
				fireUpNewProtocolFlow(pid, queue);
			}
			catch(Throwable th)
			{
				FATALEXCEPTION(th, "Uh-oh.  fireUpNewProtocolFlow() blew up.");
			}
		}
		queue.enqueue(new CommSpaceElement(ProtocolBooster.
						   LOCAL_PROTOCOL));
		queue.enqueue(new CommSpaceElement(name.
						   toCharArray()));
	}
	
	// XXX Move the following functions into a FlowManager class.

	PACKAGE static void fireUpNewAppFlow(AppConfig app)
		throws InterruptedException,
		IOException
	{
		CommSpaceHandle csh = fireUpNewFlow(
			new Credentials("boot app"), // XXX
			new ThreadPoolSpec(null, null), // XXX
			new MemSpec(0), // Does an application need UdpBuffers?
			null,
			null,
			AppBooster.class.getName(),
			app.args);
	}

	PACKAGE static void fireUpNewProtocolFlow(
			TypeID pid,
			CommQueue queue)
		throws InterruptedException,
		IOException
	{
		fireUpNewFlow(
			new Credentials("dlprotocol:" +pid), // XXX
			new ThreadPoolSpec(null, null), // XXX
			new MemSpec(100), // XXX
			pid,
			queue,
			ProtocolBooster.class.getName(),
			null);
	}

	PACKAGE static CommSpaceHandle fireUpNewFlow(
			Credentials cred,
			ThreadPoolSpec tps,
			MemSpec ms,
			TypeID pid,
			CommQueue queue,
			String className,
			String[] args)
		throws InterruptedException,
		IOException
	{
		CommSpaceHandle csh = CommSpaceHandle.newCommSpace(5);
		csh.write(0, new CommSpaceElement(getAddress()));
		csh.write(1, new CommSpaceElement(neighborsSpace));
		if(pid != null)
			csh.write(2, new CommSpaceElement(pid.unwrap()));
		if(queue != null)
			csh.write(3, new CommSpaceElement(queue));
		csh.write(4, new CommSpaceElement(unknownMidPackets.midToSpaceMap));

		FlowStartSpec fss = new FlowStartSpec(
			new ClassSpec[0], // XXX XXX XX
			className,
			args,
			csh);

		Runnable th = new Runnable() {
			public void run() {
				panic("termination handler invoked?!");
			}
		};
		
		Flow.currentFlow().newFlow(cred, tps, ms, fss, th);

		return csh;
	}
}
