/*
 * UnknownMidPackets.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.InChannelHandle;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * UnknownMidPackets is a temporary storage place for
 * packets of unknown MIDs to live in while the MID is
 * being resolved.  Although such a storage space
 * isn't essential for correctness (we could just drop
 * the packets and then have them retransmitted after
 * the MID is resolved), holding the packets should
 * smooth performance a bit.  unknownMidPackets maps
 * MIDs to arrays of BufferHandles.
 * <p>
 * In the absence of any decent synchronization mechanisms, we
 * make unknownMidPackets a lock-free data structure.
 * unknownMidPackets is a CommSpace consisting of a number
 * of smaller "spaces".	 Each space is a CommSpace with
 * the following elements:
 * <pre>
 *   0: the MID whose packets are held in this space
 *   1: the number of buffered packets in this space
 *   2, 3, ...: data items
 * </pre>
 * Each data item is either null (uninitialized),
 * a BufferHandle holding a packet, or an InChannel onto
 * which packets should be injected.  The server tries
 * to insert packets into the space, and the flow tries
 * to initialize an InChannel and fill all the data items
 * with the InChannel handle.  Whoever gets to the space
 * first simply inserts their item.  Whoever gets to the
 * space second takes the item that was already there, together
 * with their own item, so that they have both a packet and
 * an InChannel, and pushes the packet onto the InChannel.
 * An atomic swap is used to insert things into the space,
 * so that the server/flow can tell whether they got to
 * each item first or second.
 * <p>
 * This is both the client and server end (via constructor).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE class UnknownMidPackets
{
	/**
	 * Map of MID bytes to a space.  The element space
	 * has the following things stored at the following
	 * offsets:
	 * <pre>
	 *   0: the MID bytes
	 *   1: the number of buffered packets in this space
	 *   2...: the packets (one per index).
	 */
	PACKAGE final CommSpaceHandle midToSpaceMap;

	/**
	 * The next space to write a new (unknown) MID to.  This
	 * value walks through midToSpaceMap showing where to insert
	 * new MIDs.  It will wrap and start overwriting old MID maps
	 * in an (pseudo) LRU fashion.
	 */
	private int nextSpace;

	/**
	 * The number of MIDs that can be simultaneous tracked by this node.
	 */
	private final static int MIDMAPSIZE = 20;

	/**
	 * The number of packets with a given MID that are buffered
	 * before dropping.
	 */
	private final static int PACKETS_PER_SPACE = 5;

	/**
	 * This is a special flag used when "forwarding" buffered
	 * packets out of a midSpaceMap buffer.  Only used in
	 * CapsuleChannel.receive()
	 *
	 * @see CapsuleChannel#receive(BufferHandle,int)
	 */
	PACKAGE final static int NEVER_DEMULTIPLEXED = -1;

	/**
	 * Construct a server-side map for buffered packets.
	 */
	PACKAGE UnknownMidPackets()
	{
		midToSpaceMap = CommSpaceHandle.newCommSpace(MIDMAPSIZE);
		for(int i = 0; i < MIDMAPSIZE; i++)
		{
			CommSpaceHandle space =
				CommSpaceHandle.newCommSpace(
					2 + PACKETS_PER_SPACE);

			space.write(0, null);
			midToSpaceMap.write(i, new CommSpaceElement(space));
		}
	}

	/**
	 * Construct a client-side handle on a buffered packet map.
	 */
	PACKAGE UnknownMidPackets(CommSpaceHandle midToSpaceMap)
	{
		this.midToSpaceMap = midToSpaceMap;
	}

	/**
	 * Given a mid (as a byte array) find the associated CommSpace
	 * in the spaceMap.
	 * 
	 * This is deliberately not synchronized.  It is safe for
	 * concurrent execution.
	 */
	private CommSpaceHandle spaceForMID(final byte[] midBytes)
	{
		int i;
		final byte[] bytes = new byte[midBytes.length];
		final int start;

		
		if (nextSpace == 0)
			start = MIDMAPSIZE - 1;
		else
			start = (nextSpace - 1);

		/*
		 * Loop through the entire map looking for
		 * the appropriate entry.  Loop backwards starting from
		 * 'nextSpace'.
		 */
		i = start;
		do
		{
			final CommSpaceHandle space;
			final CommSpaceElement elem;
			
			assert(i >= 0);
			assert(i < MIDMAPSIZE);

			// get handle on the space at i
			space = midToSpaceMap.read(i).getCommSpaceHandle();
			
			// get the MID bytes out of the space (offset 0)
			elem = space.read(0);

			// if not empty then compare the byte arrays
			if (elem != null)
			{
				int j;
				elem.getByteArray(bytes, 0);
				for(j = 0; j < bytes.length; j++)
				{
					if(bytes[j] != midBytes[j])
						break;
				}
				if(j == bytes.length)
					return space;
			}
			
			// try previous entry
			i = (i - 1);

			if (i == -1)
				i = MIDMAPSIZE - 1;
		}
		while (i != start);
		
		return null;
	}

	/**
	 * add the given packet (which contains the given mid) to the
	 * buffered MID map.  If no packet with the MID has been
	 * seen, a new entry in the MID map is made.  Otherwise,
	 * the packet is added to the end of the appropriate list
	 * (or dropped if the maximum number of packets have already
	 * been buffered for that MID.
	 * <p>
	 * Returns true if the MID is not new (i.e., there are already
	 * packets buffered for that MID.
	 * <p>
	 * Synchronized to protect access/increment of 'nextSpace' and
	 * to prevent the same MID from being installed twice (don't know
	 * if (1) that's a real problem and (2) if its not already
	 * prevented elsewhere.)
	 * <p>
	 * XXX change to return true if packet was queued, false if
	 * dropped ??
	 */
	PACKAGE synchronized boolean addPacket(byte[] midBytes,
					       BufferHandle packet)
	{
		try
		{
			// Does the MID already have a space?
			CommSpaceHandle space;
			final boolean found;

			space = spaceForMID(midBytes);
			found = (space != null);  // the return value.

			// If the MID doesn't have a space, make one (and
			// kick out some other unlucky MID in sort of an LRU
			// fashion):
			if(!found)
			{
				// Get whatever space existed at the current index.
				CommSpaceHandle old = midToSpaceMap.read(nextSpace)
					.getCommSpaceHandle();

				// Clear it out in case it wasn't already empty.
				for(int i = 0; i < PACKETS_PER_SPACE; i++)
				{
					Object o = old.swap(2 + i, null);

					// Manually cleanse the unused BufferHandle
					// -- drop the packet.
					if(o instanceof BufferHandle)
					{
						TRACE(Trace.ANTSR_DYNLOAD, "Dropping packet in buffer " +o+ " (reusing slot in UnknownMIDMaps)");
						((BufferHandle) o).reset();
					}
				}
	
				// And put in a new space for the new MID:
				space = CommSpaceHandle.newCommSpace(
					2 + PACKETS_PER_SPACE);
				space.write(0, new CommSpaceElement(midBytes));
				space.write(1, new CommSpaceElement(0));
	
				// XXX access/incr of nextSpace needs to be protected.
				midToSpaceMap.write(nextSpace, new CommSpaceElement(space));
				nextSpace = (nextSpace + 1) % MIDMAPSIZE;
			}
	
			// Where does the next packet go in the space, and
			// is the space full?
			int nextPacket = space.read(1).getInt();
			if(nextPacket >= PACKETS_PER_SPACE)
			{
				// Drop it.
				TRACE(Trace.ANTSR_DYNLOAD, "Dropping packet in buffer " +packet+ " (Buffered MID map mid=" +midBytes+ " is full.)");
				return found;
			}

			space.write(1, new CommSpaceElement(nextPacket + 1));

			// Swap the packet in:
			BufferHandle h = new BufferHandle(Flow.currentFlow(),
				packet);
			CommSpaceElement elem = space.swap(
				2 + nextPacket,
				new CommSpaceElement(h));
			if(elem != null)
			{
				// The InChannel has already been established,
				// so go ahead and push the packet onto it.
				elem.getInChannelHandle().push(h,
					NEVER_DEMULTIPLEXED);

				// Release the packet now rather than waiting
				// for the "Clear out the old space" code
				// above to release it.
				space.write(2 + nextPacket, null);
				h.reset();
			}
			return found;
		}
		catch(Exception e)
		{
			FATALEXCEPTION(e, "Blew up adding packet to UnknownMidPacketBag");
		}
	}

	// This is deliberately not synchronized.  It is safe for
	// concurrent execution.
	PACKAGE void setInChannel(byte[] midBytes, InChannelHandle inChannel)
	{
		CommSpaceHandle space = spaceForMID(midBytes);
		if(space == null) return;

		CommSpaceElement inChan = new CommSpaceElement(inChannel);

		// Swap the InChannel in:
		for(int i = 0; i < PACKETS_PER_SPACE; i++)
		{
			CommSpaceElement elem = space.swap(
				2 + i, inChan);
			if(elem != null)
			{
				BufferHandle packet = elem.getBufferHandle();
				inChannel.push(packet, NEVER_DEMULTIPLEXED);
				packet.reset();
			}
		}
	}
}
