/*
 * UnknownPacketHandler.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.InChannel;


IMPORT_DEBUG
IMPORT_TRACE

/**
 * Handle an ANTSR packet of unknown MID
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE final class UnknownPacketHandler
	implements DeliverFunc
{
	private final Node node;
	private final DemultiplexKey classifier;
	private final java.util.Random random = new java.util.Random(); // XXX

	private static final float RANDOM_DUPMIDREQ_FRACTION = 0.05F;

	/**
	 * Create the UnknownPacketHandler.  Only need one of these
	 * per node...
	 */
	PACKAGE UnknownPacketHandler(Node node)
	{
		this(node, new ANEPDemultiplexKey(0, Capsule.ANTSRTYPE));
	}
	
	PACKAGE UnknownPacketHandler(Node node, DemultiplexKey classifier)
	{
		this.node = node;
		this.classifier = classifier;
		new InChannel(Flow.currentFlow(),
			      this.classifier,
			      Flow.currentFlow().threadPool(),
			      this);
	}

	/**
	 * Unknown packet handler.  For ANTSR packets that don't match
	 * an installed MID.
	 * <p>
	 * This is the only method defined by the DeliverFunc interface.
	 *
	 * @see DeliverFunc
	 */
	public final void receive(BufferHandle bufh, int payloadStart)
	{
		int anepHeaderLen;

		TRACE(Trace.ANTSR_DYNLOAD, "Received unknown ANTSR packet " +Debug.dbgPacketSummary(bufh, 8));

		/*
		 * By definition this receive() function only
		 * gets packets for an unknown ANTSR protocol.
		 */
		
		assert(payloadStart == 4); // just past ANEP ANTSR TypeId

		// Toss the packet if it doesn't have a mid.
		if (bufh.getLength() < (6 + 8 + 16))
		{
			// XXX log this event.
			TRACE(Trace.ANTSR_DYNLOAD, "Unknown packet is bad: length < " +(6 + 8 + 16)+ " bytes");
			return;
		}
		
		// Get the ANEP header (encoded as # of 32-bit words).
		anepHeaderLen = (((bufh.getByte(4) & 0xFF) << 8) // XXX big endian
				  | ((bufh.getByte(5) & 0xFF))) * 4;
		assert(anepHeaderLen >= (2 * 4)); // minimum allowed value

		// XXX malformed packet
		if (anepHeaderLen > bufh.getLength())
		{
			// XXX log this event
			TRACE(Trace.ANTSR_DYNLOAD, "Unknown packet is malformed packet: headerlen=" +anepHeaderLen+ "; bufh.len=" +bufh.getLength()+ ".");
			return;
		}

		// Snag the mid.  XXX duplicate code with ANTSRDemultiplexKey.matches()
		byte[] midbuf = new byte[Xdr.TYPEID];
		bufh.getBytes(anepHeaderLen + Xdr.INT, midbuf, 0, 16);
		TypeID mid = TypeID.wrap(midbuf).intern();
		assert(mid.intern() == mid);

		// XXX create a new BufferHandle that subarrays from payloadStart on?
		payloadStart = anepHeaderLen + Xdr.INT + Xdr.TYPEID;

		TRACE(Trace.ANTSR_DYNLOAD, "handling unknown mid (" +mid+ ")");

		// XXX loses payloadstart...
		boolean midAlreadySeen = PrimordialNode.unknownMidPackets
			.addPacket(mid.unwrap(), bufh);

		if(midAlreadySeen)
		{
			/*
			 * This mid request has already been sent so we don't
			 * have to worry about it since the pending request
			 * mechanism "should" take care of it.
			 */
			return;
		}

		TRACE(Trace.ANTSR_DYNLOAD, "Sending MIDIDRequest for mid (" + mid + ")");

		/*
		 * XXX this is hackish: For convenience, pull out the
		 * address of the node that this packet came from.
		 * The new flow will need the info for firing off a
		 * DLRequest capsule...
		 *
		 * Format after mid is:
		 * src (int), dst (int), and previous (int)
		 */
		int prev = bufh.getInt(payloadStart + 4 + 4);
		
		MIDIDRequestCapsule req = new MIDIDRequestCapsule(prev, mid);
		
		/*
		 * Setup the req packet to go back to whence it
		 * came.
		 */
		req.prime(node.getAddress(), 1);

		node.routeForNode(req, prev);
		/* Add a pending request for this mid request */
		node.getCache().put(mid,
				    new PendingRequest(req),
				    2);
	}

	public String toString()
	{
		return "UnknownPacketHandler for " +node;
	}
}
