/* 
 *  $Id: Xdr.Java,v 1.1 2000/01/06 01:00:52 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

IMPORT_DEBUG

/** 
 * An Xdr instance encapsulates an external, byte-oriented data
 * structure (an "External Data Representation".)  Xdr's can include
 * Strings, byte arrays, booleans, bytes, shorts, ints, longs,
 * TypeIDs.  Elements must encoded/decoded serially to/from the Xdr.
 * <p>
 * Xdr's may be hierarchical.  That is an xdr may subdivide its buffer
 * space into child buffers.
 * <p>
 * Xdr's will allocate a byte[] when constructed, or alternatively they
 * can use a given byte[], a ByteArray, another Xdr (parenting), or
 * a Core Buffer.
 * <p>
 * These methods are declared final so that they can be inlined with
 * optimization.
 *
 * @author David Wetherall
 */
public final class Xdr {

	/**
	 * The length of a boolean in an xdr.
	 */
	public final static int BOOLEAN = 1;

	/**
	 * The length of a byte in an xdr.
	 */
	public final static int BYTE = 1;

	/**
	 * The length of a short in an xdr.
	 */
	public final static int SHORT = 2;

	/**
	 * The length of a int in an xdr.
	 */
	public final static int INT = 4;

	/**
	 * The length of a long in an xdr.
	 */
	public final static int LONG = 8;

	/**
	 * The length of a typeid in an xdr.
	 */
	public final static int TYPEID = 16;

	
	/** 
	 * The byte array encoding.
	 *
	 * XXX final?
	 */
	private byte[] buf;

	/**
	 * Base offset in the buf.  The "start" of the
	 * xdr encoding.
	 *
	 * XXX final?
	 */
	private int offset;

	/**
	 * Internal buffer length.
	 *
	 * XXX final?
	 */
	private int len;

	/**
	 * Current index in the encoding.  Serializing and
	 * deserialzing automatically increment this offset.
	 */
	private int index = 0;

	/**
	 * A parent xdr.  A parent can subdivide is xdr buffer
	 * among child Xdrs.
	 *
	 * XXX final?
	 */
	private Xdr parent;


	/**
	 * Reset an xdr stream.
	 * <p>
	 * Package-access only.
	 *
	 * @return For convenience, return Xdr stream that was reset.
	 */
	PACKAGE Xdr reset()
	{
		offset = 0; // XXX is this right??
		index = 0;
		len = buf.length;
		return this;
	}

	/**
	 * Return the current index in the xdr.
	 */
	PACKAGE int index()
	{
		return index;
	}

	/**
	 * Return the number of bytes required t encode x in the
	 * xdr.  Does not actually encode the string.  Handles
	 * null strings just fine.
	 * <p>
	 * The string had best be less than 256 characters long.
	 * <p>
	 * The string had best be in ASCII. (Or some other one-to-one
	 * encoding of characters into bytes.)
	 *
	 * @param x the string to check
	 * @return the number of bytes required in the xdr
	 */
	public final static int STRING(String x)
	{
		assert(x.length() < 128);

		if (x == null)
			return BYTE;
		else
			return BYTE + x.length();
	}

	/**
	 * Return the number of bytes required to encode a
	 * ByteArray in the xdr.  Does not actually encode
	 * the ByteArray.  Handles null just fine.
	 * <p>
	 * The ByteArray must be less than 65,535 bytes.
	 *
	 * @param x the ByteArray to check
	 * @return the number of bytes required in the xdr.
	 */
	public final static int BYTEARRAY(ByteArray x)
	{
		assert(x.len < ((1 << 16)-1));

		if (x == null)
			return SHORT;
		else
			return SHORT + x.len;
	}

	/**
	 * Return the size of the buffer used by this xdr.
	 *
	 * @return the size of the buffer used by this xdr.
	 */
	public final int length()
	{
		return len;
	}

	/**
	 * Encode the given boolean into this xdr.
	 * There must be sufficient space in this
	 * xdr preallocated.
	 */
	public final void PUT(boolean x)
	{
		if (index + BOOLEAN > len)
			throw new IndexOutOfBoundsException("Not enough space to encode boolean.");
		int j = offset + index;
		if (x)
			buf[j] = (byte)0xff;
		else
			buf[j] = (byte)0;
		index += BOOLEAN;
	}

	public final void PUT(byte x) {
		if (index + BYTE > len)
			throw new IndexOutOfBoundsException("Not enough space to encode byte: " +x);
		int j = offset + index;
		buf[j] = x;
		index += BYTE;
	}

	public final void PUT(short x) {
		if (index + SHORT > len)
			throw new IndexOutOfBoundsException("Not enough space to encode short: " +x);
		int j = offset + index;
		buf[j++] = (byte)(x>>>8);
		buf[j] = (byte)x;
		index += SHORT;
	}

	public final void PUT(int x) {
		if (index + INT > len)
			throw new IndexOutOfBoundsException("Not enough space to encode int: " +x);
		int j = offset + index;
		buf[j++] = (byte)(x>>>24);
		buf[j++] = (byte)(x>>>16);
		buf[j++] = (byte)(x>>>8);
		buf[j] = (byte)x;
		index += INT;
	}

	public final void PUT(long x) {
		if (index + LONG > len)
			throw new IndexOutOfBoundsException("Not enough space to encode long: " +x);
		int j = offset + index;
		buf[j++] = (byte)(x>>>56);
		buf[j++] = (byte)(x>>>48);
		buf[j++] = (byte)(x>>>40);
		buf[j++] = (byte)(x>>>32);
		buf[j++] = (byte)(x>>>24);
		buf[j++] = (byte)(x>>>16);
		buf[j++] = (byte)(x>>>8);
		buf[j] = (byte)x;
		index += LONG;
	}

	public final void PUT(TypeID t) {
		if (index + TYPEID > len)
			throw new IndexOutOfBoundsException("Not enough space to encode TypeID.");
		int j = offset + index;
		System.arraycopy(t.unwrap(), 0, buf, j, TYPEID);
		index += TYPEID;
	}

	public final void PUT(String x) {
		if (index + STRING(x) > len)
			throw new IndexOutOfBoundsException("Not enough space to encode String: " +x);
		int j = offset + index;
		if (x == null)
		{
			buf[j] = (byte)0;
			index += BYTE;
		}
		else
		{
			int l = x.length();
			buf[j++] = (byte)l;
			byte [] xbytes = x.getBytes(); 
			assert(xbytes.length == x.length()); // XXX fails for uni-code..
			System.arraycopy(xbytes, 0, buf, j, l);
			// deprecated: x.getBytes(0, l, buf, j);
			index += BYTE + l;
		}    
	}

	public final void PUT(ByteArray x) {
		if (index + BYTEARRAY(x) > len)
			throw new IndexOutOfBoundsException("Not enough space to encode ByteArray.");
		int j = offset + index;
		if (x == null)
		{
			buf[j++] = (byte)0; 
			buf[j] = (byte)0;
			index += SHORT;
		}
		else
		{
			int l = x.len;
			buf[j++] = (byte)(l>>>8);
			buf[j++] = (byte)l;
			if (x.buf != buf || x.offset != j)
				System.arraycopy(x.buf, x.offset, buf, j, l);
			index += SHORT + l;
		}
	}

	public final void PUT(byte[] x, int start, int length) {
		if (index + length > len)
			throw new IndexOutOfBoundsException("Not enough space to encode byte[].");
		int j = offset + index;
		System.arraycopy(x, start, buf, j, length);
		index += length;
	}

	public final boolean BOOLEAN() {
		if (index + BOOLEAN > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += BOOLEAN;
		if (buf[j] != 0)
			return true;
		else
			return false;
	}

	public final byte BYTE() { 
		if (index + BYTE > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += BYTE;
		return (byte)(buf[j]&0xff);
	}
  
	public final short SHORT() {
		if (index + SHORT > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += SHORT;
		return (short)((buf[j++]&0xff)<<8 | (buf[j]&0xff));
	}

	public final int INT() {
		if (index + INT > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += INT;
		return (buf[j++]&0xff)<<24 | (buf[j++]&0xff)<<16 |
			(buf[j++]&0xff)<<8 | buf[j]&0xff;
	}

	public final long LONG() {
		if (index + LONG > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += LONG;
		return ((long)buf[j++]&0xff)<<56 | ((long)buf[j++]&0xff)<<48 |
			((long)buf[j++]&0xff)<<40 | ((long)buf[j++]&0xff)<<32 |
			((long)buf[j++]&0xff)<<24 | ((long)buf[j++]&0xff)<<16 |
			((long)buf[j++]&0xff)<<8 | (long)buf[j]&0xff;
	}

	public final TypeID TYPEID() { 
		if (index + TYPEID > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		index += TYPEID;
		return new TypeID(buf, j);
	}

	public final String STRING() {
		if (index + SHORT > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		int l = buf[j++];
		if (index + BYTE + l > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		index += BYTE + l;
		if (l == 0)
			return null;
		else 
			return new String(buf, j, l);
	}

	public final ByteArray BYTEARRAY() { 
		if (index + SHORT > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		int l = (buf[j++]&0xff)<<8 | (buf[j++]&0xff);
		if (index + SHORT + l > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		index += SHORT;
		ByteArray data = null;
		if (l != 0)
			data = new ByteArray(buf, offset + index, l);
		index += l;
		return data;
	}

	public final void BYTES(byte[] x, int start, int length) {
		if (index + length > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		int j = offset + index;
		System.arraycopy(buf, j, x, start, length);
		index += length;
	}

	public final void skip(int count) {
		if (index + count > len)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		index += count;
	}

	PACKAGE byte[] getBytes()
	{
		assert(offset == 0);
		return buf;
	}

	/**
	 * Create a new xdr with a buffer of len
	 * bytes for encoding.
	 *
	 * @param len the size of the buffer to create.
	 */
	public Xdr(int len)
	{
		buf = new byte[len];
		offset = 0;
		this.len = len;
	}

	/**
	 * Create a new xdr using the given buffer.
	 * Does NOT copy the given buf.
	 * <p>
	 * XXX probably shouldn't be public.
	 *
	 * @param buf the buffer to use.
	 */
	public Xdr(byte[] buf)
	{
		this(buf, 0, buf.length);
	}

	/**
	 * Create a new xdr using the buffer in the given ByteArray.
	 *
	 * @param buf the ByteArray to steal the byte[] from
	 * @param offset offset to start at in the ByteArray
	 */
	public Xdr(ByteArray buf, int offset)
	{
		this(buf, offset, buf.len - offset);
	}
	
	/**
	 * Create a new xdr using the buffer in the given ByteArray.
	 *
	 * @param buf the ByteArray to steal the byte[] from
	 * @param offset offset to start at in the ByteArray
	 * @param len length of ByteArray to use
	 */
	public Xdr(ByteArray buf, int offset, int len)
		throws IndexOutOfBoundsException
	{
		if (offset + len > buf.len || offset < 0 || len < 0)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		this.buf = buf.buf;
		this.offset = buf.offset + offset;
		this.len = len;
	}

	/**
	 * Create a new xdr using the given portion of the given buffer.
	 *
	 * @param buf the buffer to use.
	 * @param offset only encode in the portion of the buffer starting at offset.
	 */
	public Xdr(byte[] buf, int offset)
	{
		this(buf, offset, buf.length - offset);
	}

	/**
	 * Create a new xdr using the given portion of the given buffer.
	 *
	 * @param buf the buffer to use.
	 * @param offset only encode in the portion of the buffer starting at offset.
	 * @param len only encode using len bytes of the buffer.
	 */
	public Xdr(byte[] buf, int offset, int len) 
		throws IndexOutOfBoundsException
	{
		if (offset + len > buf.length || offset < 0 || len < 0)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		this.buf = buf;
		this.offset = offset;
		this.len = len;
	}
  
	/**
	 * Package-only constructor for creating an xdr 
	 * as a subsequence of a parent xdr.  New xdr
	 * will manage remainder of parent xdr's buffer.
	 *
	 * @param parent Xdr to steal buffer space from.
	 * @param offset Offset in parent buffer to start.
	 */
	PACKAGE Xdr(Xdr parent, int offset)
	{
		this(parent, offset, parent.len-offset);
	}

	/**
	 * Package-only constructor for creating an xdr 
	 * as a subsequence of a parent xdr.  New xdr
	 * will manage only the given portion of parent's buffer.
	 *
	 * @param parent Xdr to steal buffer space from.
	 * @param offset Offset in parent buffer to start.
	 * @param len Length of the buffer to manage.
	 */
	PACKAGE Xdr(Xdr parent, int offset, int len)
		throws IndexOutOfBoundsException
	{
		if (offset < 0 || offset + len > parent.len || len < 0)
			throw new IndexOutOfBoundsException("No more data left in xdr.");
		this.parent = parent;
		buf = parent.buf;
		this.offset = parent.offset + offset;
		this.len = len;
	}
}
