/*
 * DynamicRouteTable.Java
 * 
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package JANOS_ANTSR_ROUTE_PKG;

import JANOS_ANTSR_CORE_PKG.Application;
import JANOS_ANTSR_CORE_PKG.Capsule;
import JANOS_ANTSR_CORE_PKG.Node;
import JANOS_ANTSR_CORE_PKG.Xdr;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;

import JANOS_IO_PKG.Log;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_NODEOS_PKG.AddressFormatException;

import java.util.Vector; // XXX
import java.util.Hashtable; // XXX
import java.util.Enumeration; // XXX

IMPORT_TRACE
IMPORT_DEBUG


/**
 * The "app" that maintains a node's dynamic routing tables.
 * <p>
 * This is really just an app so it can justify sending capsules off
 * without priming them off some other capsule.  (All this code used
 * to be part of DynamicRouteTable.)
 * <p>
 * Understands the default '-port' argument that all apps grok.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 * @author David Wetherall
 */
public class DynamicRouteApp
	extends Application
{
	public static final int APP_PORT = 13;

	private static final String[] defaults = {"-port", Integer.toString(APP_PORT)};

	/**
	 * Maximum number of route updates to stuff into a
	 * single RouteUpdateCapsule.  (Each update costs 5 bytes).
	 */
	private final static int MAXROUTESPERUPDATE = 64;

	/**
	 * The table this app is keeping up to date.
	 */
	private DynamicRouteTable table;

	/**
	 * The node this table is maintained on.
	 */
	private Node node;

	/**
	 * Used for suppressing triggered updates.  Triggered
	 * updates are scheduled when the routing table is
	 * changed.  Once a triggered update is scheduled,
	 * futher updates are suppressed until the trigger
	 * update is executed.
	 */
	private boolean suppressTriggerEvents;
	
	/**
	 * The queue of timers and events that cause updates.
	 */
	private final DynamicRouteEvents events;

	/**
	 * App constructors can't do anything useful.  Ideally,
	 * I would like to set this.node and this.table, but I
	 * can't because the node hasn't been plugged into this
	 * app yet, and I need to node to get the table...
	 */
	public DynamicRouteApp()
	{
		TRACE(Trace.ANTSR_ROUTE_UPDATE, "DynamicRouteApp starting.");

		this.events = new DynamicRouteEvents();
		this.suppressTriggerEvents = false;
	}

	/**
	 * Parse command line arguments.
	 * <p>
	 * See the class description for a list of the arguments and
	 * their meaning.
	 */
	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);
		super.setArgs(k);
	}

	/**
	 * Initialize the event queue.
	 */
	private void initEvents()
	{
		// XXX the only installed routes should be via gateways...
		// see DynamicRouteTable.addRoute(...)
		long now = node.time();

		for (Enumeration e = table.getGateways();
		     e.hasMoreElements(); /* */)
		{
			DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();
			events.setExpireTimer(r, now);
		}
	}
	
	/**
	 * Do initialization of the App.
	 * <p>
	 * XXX one-time function.
	 */
        private void initApp()
        {
		this.node = thisNode();
		assert(this.node);

		// XXX is the type cast a sufficient run-time check?
		this.table = (DynamicRouteTable)node.getRouteTable(); // XXX XXX XXX
		assert(this.table);

		TRACE(Trace.ANTSR_ROUTE_UPDATE, "DynamicRouteApp starting.");

		/*
		 * Install the update protocol (register capsules, etc).
		 */
		try
		{
			DynamicRouteProtocol rp = new DynamicRouteProtocol();
			TRACE(Trace.ANTSR_ROUTE_UPDATE, "Installing " +rp);
			node.register(rp);
		}
		catch (Exception e)
		{
			FATALEXCEPTION(e, "DynamicRouteProtocol setup failure.");
		}

		/*
		 * Initialize events from the gateways setup in the routing table
		 */
		initEvents(); // XXX one-time function
	}

	/**
	 * The active route table logic.
	 * <p>
	 * Harmless if invoked on a static table.
	 */
	public void start()
	{
		/*
		 * Setup local state that couldn't be set in the constructor.
		 */
	        initApp();

		// Synchronize on this table.  Is that sufficient??  
		// Drops lock in table.wait, below...
		synchronized (table)
		{
			// Install next periodic wakeup
			events.setPeriodicTimer(node.time());
			
			while (true) 
			{
				final DynamicRouteEvent event;

				assert(!events.isEmpty());

				// find the next scheduled event
				event = events.findFirstTimer();
				
				// if it's time, then process it
				long delay = event.when - System.currentTimeMillis();
				if (delay <= 0)
				{
					// sanity check for leftovers in the queue
					assert(event.what != DynamicRouteEvent.EMPTY); // XXX
					
					// Each event may have more than one duty.  Do all that apply.

					if ((event.what & DynamicRouteEvent.GC) != 0) {
						TRACE(Trace.ANTSR_ROUTE_TIMER, "GC Event");
						gcEntries(event);
					}
					
					if ((event.what & DynamicRouteEvent.EXPIRE) != 0) {
						TRACE(Trace.ANTSR_ROUTE_TIMER, "EXPIRE Event");
						
						expireEntries(event);
					}
					
					// Note that route messages are sent if the event is
					// a periodic or trigger timer.  Don't waste work.
					if ((event.what & DynamicRouteEvent.PERIODIC) != 0) {
						TRACE(Trace.ANTSR_ROUTE_TIMER, "PERIODIC Event");
						
						sendRouteMessages(DynamicRouteEvent.PERIODIC);
						events.setPeriodicTimer(event.when);
						
					} else if ((event.what & DynamicRouteEvent.TRIGGER) != 0) {
						TRACE(Trace.ANTSR_ROUTE_TIMER, "TRIGGER Event");
						
						sendRouteMessages(DynamicRouteEvent.TRIGGER);
						clearChangeFlags();
					}
					
					// clear trigger even if we had a PERIODIC+TRIGGER timer.
					if ((event.what & DynamicRouteEvent.TRIGGER) != 0)
						suppressTriggerEvents(false);

					// Toss the event, we've handled it.
					events.clearTimer(event);
				}
				else
				{
					/* 
					 * First event is far in the
					 * future.  Block until that day arrives.
					 */
					try
					{
						TRACE(Trace.ANTSR_ROUTE_TIMER, "RouteTable: Waiting for " + delay + "ms ...");

						// *wait* (not sleep) so that we release the lock on this.
						table.wait(delay);
					}
					catch (InterruptedException e)
					{
						// Spurious wakeups are harmless
						// XXX statcount this
						;
					}
				}
			}
			
			// NOT REACHED
		}
	}


	// -- BEGIN LOCAL ROUTE TABLE PROCESSING ---

	/**
	 * Walks all routes and moves expired routes into the gcPool.
	 */
	private void expireEntries(DynamicRouteEvent e)
	{
		boolean change;
		do // XXX why loop until nothing changes?  Always true on second iteration?
		{
			change = false;
			for (Enumeration f = table.getRoutes(); f.hasMoreElements();) {
				DynamicRouteEntry r = (DynamicRouteEntry)f.nextElement();
				if (r.getExpiration() == e.when)
				{
					TRACE(Trace.ANTSR_ROUTE_UPDATE, "RouteTable: Route to "+
						    Address.toString(r.getLogicalDest()) +" (gw="+
						    Address.toString(r.getLogicalNext()) + ") expired.");
					
					// Move route to GC Pool.
					r.setCost(table.INFINITY);
					table.gcPool().put(new Integer(r.getLogicalDest()), r);
					table.removeRoute(r);
					r.change = true;
					events.setGCTimer(r, r.getExpiration());

					scheduleTrigger(r.getExpiration());
					change = true;

					break;
				}
			}
		} while (change);
	}
	
	/**
	 * Delete all routes in the gcPool that have expired.  Such routes
	 * are lost for all time.
	 */
	private void gcEntries(DynamicRouteEvent e)
	{
		boolean change;
		
		do // XXX why is this loop here?
		{
			change = false;
			for (Enumeration f = table.gcPool().elements(); f.hasMoreElements();)
			{
				DynamicRouteEntry r = (DynamicRouteEntry)f.nextElement();
				if (r.getExpiration() == e.when) {
					TRACE(Trace.ANTSR_ROUTE_UPDATE, "RouteTable: GCing route to " + Address.toString(r.getLogicalDest()));
					table.gcPool().remove(new Integer(r.getLogicalDest()));
					change = true;
					break;
				}
			}
		} while (change);
	}
	
	/**
	 * Mark all routes as 'unchanged'.
	 */
	private void clearChangeFlags()
	{
		for (Enumeration f = table.getRoutes(); f.hasMoreElements();)
		{
			DynamicRouteEntry d = (DynamicRouteEntry)f.nextElement();
			d.change = false;
		}
		
		for (Enumeration f = table.gcPool().elements(); f.hasMoreElements();)
		{
			DynamicRouteEntry d = (DynamicRouteEntry)f.nextElement();
			d.change = false;
		}
	}

	/**
	 * Send a route update to the given node.
	 * <p>
	 * XXX this is completely out of synch with the new world order.
	 */
	private void sendUpdate(DynamicRouteEntry g, Vector r)
	{
		boolean sentIt;
		DynamicRouteCapsule cap = 
			new DynamicRouteCapsule(g.getLogicalDest(),
						node.getPhysicalAddress(),
						r);
		
		// XXX may cause recursion into the
		// RouteTable.. shouldn't be bad, though.
		sentIt = send(cap);

		if (!sentIt)
		{
			node.log(Log.WARNING, "Failed to send DynamicRouteCapsule " +cap+ ".  That's bad, m'kay.");
		}
		
	}
	
	/**
	 * Send a message to a host about a route.  The given
	 * host should be a gateway host for this node.
	 */
	private void sendRouteWorker(DynamicRouteEntry gatewayRoute, int cause)
	{
		final Vector routeChunk = new Vector( MAXROUTESPERUPDATE ); // XXX hit that limit often?
		boolean update = false;
		boolean gcswitch = false;
		int routes = 0;
		int poison = 0;
		boolean gwchange = false;
		
		for (Enumeration f = table.getRoutes(); f.hasMoreElements();)
		{
			DynamicRouteEntry currRoute = (DynamicRouteEntry)f.nextElement(); // xxx check cast
			
			// search the gc pool after the main table... oi.
			if (!gcswitch & !f.hasMoreElements())
			{
				f = table.gcPool().elements();
				gcswitch = true;
			}
			
			// skip the route to that gateway.  Record a route
			// change to the gateway
			if (currRoute.getLogicalDest() == gatewayRoute.getLogicalDest())
			{
				if (currRoute.change)
					gwchange = true;
				continue;
			}
			
			// Don't spew unchanged route on a trigger event
			if ((cause == DynamicRouteEvent.TRIGGER)
			    && (!currRoute.change))
				continue;

			// Route is changed or this is a periodic update.
			
			// Poison all routes through the current gateway.  It knows
			// better than we do the cost of the path... 
			int cost = currRoute.getCost();
			if (currRoute.getLogicalNext() == gatewayRoute.getLogicalDest())
			{
				cost = table.INFINITY;
				poison++;
			}
			
			DynamicRouteEntry u = new DynamicRouteEntry(currRoute);
			u.setCost(cost);
			routeChunk.addElement(u);
			routes++;
			
			// If we've hit high-water mark, send what we've got and continue afresh.
			if (routes == MAXROUTESPERUPDATE) {
				
				TRACE(Trace.ANTSR_ROUTE_UPDATE, "RouteTable: Sending " + routes + " routes ("+ poison + " poison) to " + Address.toString(gatewayRoute.getLogicalDest()));
				
				sendUpdate(gatewayRoute, routeChunk);
				routeChunk.removeAllElements();
				routes = 0;
				poison = 0;
				update = true;
			}
		}
		
		// Send final chunk of route updates.  For periodic
		// timers, make sure we sent at least one update (even if its empty).
		// Or, if the gateway changed on a trigger, send an empty update.
		if ((routes > 0)
		    || (!update && (cause == DynamicRouteEvent.PERIODIC))
		    || (gwchange && (cause == DynamicRouteEvent.TRIGGER)))
		{
			TRACE(Trace.ANTSR_ROUTE_UPDATE, "RouteTable: Sending " + routes + " routes ("+ poison + " poison) to " + Address.toString(gatewayRoute.getLogicalDest()));
			
			sendUpdate(gatewayRoute, routeChunk);
			routeChunk.removeAllElements();
			routes = 0;
			poison = 0;
		}
	}
	
	/**
	 * Send a message to each gateway and to host which
	 * has a cost of "1" to reach.  For each gateway send
	 * a long list of route entries.
	 * <p>
	 * XXX this is O(N * N) when N is the *number of routes*!
	 *
	 * @param cause the reason this was invoked (Trigger or periodic).
	 */
	private void sendRouteMessages(int cause)
	{
		assert((cause == DynamicRouteEvent.TRIGGER)
		       || (cause == DynamicRouteEvent.PERIODIC));

		// generate messages for each directly connected gateway
		for (Enumeration e = table.getGateways();
		     e.hasMoreElements(); /* */)
		{
			DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();      
			sendRouteWorker(r, cause);
		}
		
//		  for (Enumeration e = table.getRoutes();
//		       e.hasMoreElements(); /* */)
//		  {
//			  DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();      
//			  if ((r.getCost() == 1)
//			      && (gateways.get(new Integer(r.getLogicalDest())) == null))
//				  sendRouteWorker(r, cause);
//		  }
	}

	/**
	 * Receive and process RouteUpdateCapsule.
	 * RouteUpdateCapsules contain zero or more route update
	 * records.  Each "record" consists of a destination
	 * and a "distance" to that destination.
	 *
	 * @param unknownCap the RouteUpdateCapsule.
	 */
	public void receive(Capsule unknownCap) {
		DynamicRouteCapsule cap;

		if (!(unknownCap instanceof DynamicRouteCapsule))
			return;

		cap = (DynamicRouteCapsule)unknownCap;
		
		TRACE(Trace.ANTSR_ROUTE_UPDATE, "Got an update capsule " +cap);

		/*
		 * Note that update capsules should *never* have
		 * to forward themselves because they're only sent
		 * to 1-hop-away gateways.
		 */
		if (node.getAddress() != cap.getDst())
		{
			node.log(Log.WARNING,
				 "DynamicRouteApp received a DynamicRouteCapsule destined for"
				 + Address.toString(node.getAddress())
				 + " instead of "
				 + Address.toString(cap.getDst())
				 + "(me).  Dropping capsule.");
			return;
		}

		// XXX ??
		if (!(table instanceof DynamicRouteTable))
			return;

		boolean changed = false;
		long now = node.time();
		
		TRACE(Trace.ANTSR_ROUTE_UPDATE, "RouteUpdateCapsule: Arrived from " + Address.toString(cap.getSrc()) + " ("+ cap.nroutes + " routes)");
		
		/*
		 * Synchronize on the table.  We also synch on the table
		 * up in the start() method.
		 */
		synchronized(table)
		{ 
			DynamicRouteEvent n1 = events.findFirstTimer();
			
			if (newRouteInfo(cap.getSrc(), 0, cap.getSrc(), cap.srcPhysAddr, now))
				changed = true;
			
			if (cap.nroutes > 0)
			{
				Xdr xdr = new Xdr(cap.routes, 0);
				
				for (int i = 0; i < cap.nroutes; i++) {
					int dest = xdr.INT();
					int dist = xdr.BYTE();
					
					if (newRouteInfo(dest, dist, cap.getSrc(), cap.srcPhysAddr, now))
						changed = true;
				}
			}
			
			if (changed)
			{
				// schedule a trigger update
				scheduleTrigger(now);
			}

			// wake the scheduler if we've altered the next event
			DynamicRouteEvent n2 = events.findFirstTimer();
			if (n2.when < n1.when)
				table.notify();
		}
		return;
	}

	/**
	 * Provide new information for a single route.  If the info
	 * is useful, update the route table.  It is the caller's
	 * responsibility to schedule a trigger update if there are
	 * changes to the route table.
	 * <p>
	 * Invoked by exclusivly by <code>RouteUpdateCapsule.evalutate</code>.
	 *
	 * @param dest the active address of the destination to update a route to
	 * @param dist the distance to that node from the gateway
	 * @param gw the active address of the gateway (the "next hop" to dest)
	 * @param gwPhysAddr the physical address of the gateway
	 * @param now what time is it my dear?
	 * @return true if a change was made; false if not.
	 * @see RouteUpdateCapsule#evaluate(Node)
	 */
	PACKAGE boolean newRouteInfo(int dest, int dist,
				     int gw, String gwPhysAddr,
				     long now)
	{
		DynamicRouteEntry oldRoute = (DynamicRouteEntry)table.getRoute(dest); // XXX check for exception
		int cost = dist + 1;

		TRACE(Trace.ANTSR_ROUTE_UPDATE, "Updated route: to " +Address.toString(dest)+ " via " + Address.toString(gw) + "(" +gwPhysAddr + ")");

		// Snap all costs at infinity.
		if (cost > table.INFINITY)
			cost = table.INFINITY;
		
		if (oldRoute == null)
		{
			DynamicRouteEntry newRoute;

			// Bail on the expensive ones...
			if (cost == table.INFINITY)
				return false;
			
			try
			{
				newRoute = new DynamicRouteEntry(dest, gw, gwPhysAddr);
				newRoute.setCost(cost);
				newRoute.change = true;
				
			}
			catch (NoSuchNodeAddress nsne)
			{
				node.log(Log.WARNING, "Route update with bad address: " +nsne.getMessage());
				return false;
			}
			catch (AddressFormatException afe)
			{
				node.log(Log.WARNING, "Route update with bad address: " +afe.getMessage());
				return false;
			}

			// might have to resuscitate from gc pool
			DynamicRouteEntry gc = (DynamicRouteEntry)table.gcPool().get(new Integer(dest));
			if (gc != null)
			{
				events.clearGCTimer(gc.getExpiration());
				table.gcPool().remove(new Integer(dest));
			}

			TRACE(Trace.ANTSR_ROUTE_UPDATE, "Adding as a new route: " +newRoute);
			
			table.addRoute(newRoute);
			events.setExpireTimer(newRoute, now);
			
			// Made changes
			return true;
		}
		// route exists, see if anything changed as per input processing spec
		else if (((oldRoute.getLogicalNext() == gw) && (cost != oldRoute.getCost()))
			 || (cost < oldRoute.getCost()))
		{
			DynamicRouteEntry newRoute;

			try
			{
				newRoute = new DynamicRouteEntry(dest, gw, gwPhysAddr);
				newRoute.setCost(cost);
				newRoute.change = true;
			}
			catch (NoSuchNodeAddress nsne)
			{
				node.log(Log.WARNING, "Route update with bad address: " +nsne.getMessage());
				return false;
			}
			catch (AddressFormatException afe)
			{
				node.log(Log.WARNING, "Route update with bad address: " +afe.getMessage());
				return false;
			}

			
			// Found in live pool, cannot be in gc pool
			assert(table.gcPool().get(new Integer(dest)) == null);

			/*
			 * XXX new world order: cannot update a RouteEntry.  
			 * (Specifically, if the gw has changed).  Thus,
			 * have to remove and re-insert.
			 */
			
			TRACE(Trace.ANTSR_ROUTE_UPDATE, "Updating an old route: " +newRoute);

			events.clearExpireTimer(oldRoute.getExpiration());
			table.removeRoute(oldRoute);
			// drop oldRoute on the floor...
			
			if (cost == table.INFINITY)
			{
				// infinite links go right into the gc pool
				table.gcPool().put(new Integer(dest), newRoute);
				newRoute.change = true;
				events.setGCTimer(newRoute, newRoute.getExpiration());
			}
			else
			{
				table.addRoute(newRoute);
				events.setExpireTimer(newRoute, now);
			}
			
			
			return true;
		}      
		
		// finally, just reset the timers if confirming update
		else if ((oldRoute.getLogicalNext() == gw)
			 && (cost == oldRoute.getCost()))
		{
			// Found in live pool, cannot be in gc pool
			assert(table.gcPool().get(new Integer(dest)) == null);

			events.clearExpireTimer(oldRoute.getExpiration());
			events.setExpireTimer(oldRoute, now);
		}
			
		// if we reach here, there were no changes
		return false;
	}

	// -- END ROUTEUPDATECAPSULE INTERFACES --
	
					
	/**
	 * Suppress scheduling of further trigger events.  Or
	 * don't.
	 */
	PACKAGE void suppressTriggerEvents(boolean doit)
	{
		suppressTriggerEvents = doit;
	}

	/**
	 * Schedule a trigger event for some time after 'when'.
	 * If triggers are surpressed (because one has already been
	 * scheduled but not executed) the request is dropped on the floor.
	 * <p>
	 * Invoked in here and by RouteUpdateCapsule.
	 */
	PACKAGE void scheduleTrigger(long when)
	{
		if (!suppressTriggerEvents)
		{
			events.setTriggerTimer(when);
			suppressTriggerEvents = true;
		}
	}
}
