/* 
 *  $Id: DynamicRouteCapsule.Java,v 1.2 2000/01/12 21:36:27 stack Exp $
 *  
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_ROUTE_PKG;

import JANOS_NODEOS_PKG.Address;

import JANOS_IO_PKG.Log;

import JANOS_ANTSR_CORE_PKG.ByteArray;
import JANOS_ANTSR_CORE_PKG.Capsule;
import JANOS_ANTSR_CORE_PKG.Node;
import JANOS_ANTSR_CORE_PKG.Xdr;

import JANOS_ROUTE_PKG.RouteTable;
import JANOS_ROUTE_PKG.RouteEntry;

import java.util.Vector; // XXX
import java.util.Enumeration; // XXX

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Route update capsule.
 * <p>
 * Used to be a built-in capsule type.  Now it is not.
 * <p>
 * XXX fix the PACKAGE-accessible fields.  (Need to be accessible in DynamicRouteApp.)
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 * @author David Wetherall
 */
public class DynamicRouteCapsule
	extends Capsule
{
	final private static byte[] MID = findMID(JANOS_ANTSR_ROUTE_PKG_S + ".DynamicRouteCapsule");
	protected byte[] mid()
	{
		return MID;
	}
	
	final private static byte[] PID = findPID(JANOS_ANTSR_ROUTE_PKG_S + ".DynamicRouteCapsule");
	protected byte[] pid()
	{
		return PID;
	}
	
	/**
	 * Physical address of the node this DynamicRouteCapsule originated at.
	 * <p>
	 * As a String because only the Core is allowed to parse these things.
	 * XXX not very safe for pushing thorugh the network, now is it?
	 */
	PACKAGE String srcPhysAddr;
	
	/**
	 * Number of route updates this capsule contains.  Controls
	 * the size of the <code>routes</code> array.
	 * <p>
	 * XXX this need not be encoded/decoded.  It can be inferred
	 * from the size of the byte array (which is already encoded
	 * for us by the Xdr routines).
	 */
	PACKAGE int nroutes;

	/**
	 * The route updates.  Each update consists of:
	 * <pre>
	 * logical destination (INT)
	 * cost (byte)
	 * </pre>
	 */
	PACKAGE ByteArray routes;
	
	/**
	 * The length of this capsule.  Adds the length of the
	 * <code>physAddr</code> string, 4 bytes for the
	 * <code>nroutes</code> word, and zero or more bytes for the
	 * <code>routes</code> array.
	 */
	public int length()
	{
		int s = Xdr.STRING(srcPhysAddr) +
			Xdr.INT + // nroutes
			((nroutes > 0)
			 ? Xdr.BYTEARRAY(routes)
			 : 0);
		return super.length() + s;
	}
	
	public Xdr encode() 
	{
		Xdr xdr = super.encode();
		xdr.PUT(srcPhysAddr);
		xdr.PUT(nroutes);
		if (nroutes > 0)
			xdr.PUT(routes);
		return xdr;
	}
	
	public Xdr decode()
	{
		Xdr xdr = super.decode();
		srcPhysAddr = xdr.STRING();
		nroutes = xdr.INT();
		if (nroutes > 0)
			routes = xdr.BYTEARRAY();
		else
			routes = null;
		return xdr;
	}
	
	/**
	 * When an update capsule arrives at its destination Node, it
	 * gets delivered to the app that maintains the routing table.
	 */
	public boolean evaluate(Node node)
	{
		if( node.getAddress() == getDst() )
		{
			return node.deliverToApp(this,
						 DynamicRouteApp.APP_PORT);
		}
		else
			return node.routeForNode(this, getDst());
	}  
	
	public String toString()
	{
		return "[" + super.toString()
			+ "; sphys='" + srcPhysAddr + "'"
			+ "; nr=" + nroutes + "]";
	}
	

	/**
	 * Construct an empty DynamicRouteCapsule.  This is only invoked
	 * when a node receives a DynamicRouteCapsule.
	 */
	public DynamicRouteCapsule()
	{
	}
	
	/**
	 * Create a DynamicRouteCapsule to send to a remote host.
	 * This capsule will be sent to the activeDest host.
	 * <p>
	 * It contains the *physical address* of the host that sends
	 * it.  (Because, hey, that's what routing does.)
	 * <p>
	 * Lastly, it will contain a suite of DynamicRouteEntries for the
	 * destination host to store in its RouteTable.
	 *
	 * @param activeDest the destination for this capsule.  Must be 1 hop away.
	 * @param srcPhysAddr the physical address of the the sender.  As a string.
	 * @param routeVec a list of updated route entries for the remote node
	 */
	public DynamicRouteCapsule(int activeDest,
				   String srcPhysAddr,
				   Vector routeVec)
	{
		this.setDst(activeDest);
		this.srcPhysAddr = srcPhysAddr;
		this.nroutes = routeVec.size();

		if (nroutes > 0)
		{
			routes = new ByteArray(nroutes * (Xdr.INT+Xdr.BYTE));
			Xdr xdr = new Xdr(routes, 0);
			for (Enumeration e = routeVec.elements(); e.hasMoreElements();)
			{
				DynamicRouteEntry re = (DynamicRouteEntry)e.nextElement(); // XXX checkcast
				xdr.PUT(re.getLogicalDest());
				xdr.PUT((byte)re.getCost());
			}
		}
	}
}
