/*
 * DynamicRouteTable.Java
 * 
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package JANOS_ANTSR_ROUTE_PKG;

import JANOS_ANTSR_CORE_PKG.Node;
// import JANOS_ANTSR_CORE_PKG.RouteTable;

import JANOS_ROUTE_PKG.WriteableRouteTable;
import JANOS_ROUTE_PKG.RouteTable;
import JANOS_ROUTE_PKG.RouteEntry;

import JANOS_IO_PKG.Log;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_NODEOS_PKG.AddressFormatException;

import java.io.IOException; // XXX
 
import java.util.Hashtable; // XXX
import java.util.Enumeration; // XXX

IMPORT_TRACE
IMPORT_DEBUG

/**
 * The ANTSR DynamicRouteTable contains all of the state and code
 * for maintaining a dynamic routing table.  This includes
 * sending out periodic updates, sending out immediate updates,
 * (and receiving both).  GC'ing stale routes, etc.
 * <p>
 * I think there are some problems in here for a node that has
 * more than one physical address.  It will probably send the
 * wrong info out to the other nodes...
 * <p>
 * Based very strongly on the dynamic route table from the original
 * MIT ANTS.  I renamed and moved a lot of things around, but
 * I (hopefully) didn't change the update algorithm.
 * <p>
 * XXX access control issues.
 */
public class DynamicRouteTable
	extends WriteableRouteTable
	implements JANOS_ANTSR_CORE_PKG.RouteTable
{
	/**
	 * Maximum cost of travelling to a node.  Dunno why 16 is a
	 * good number.
	 */
	public final static int INFINITY = 16;
	
	/**
	 * GC pool; unavailable for retrieving routes from.
	 * <p>
	 * DynamicRouteEntries are added to the GC pool when they "expire".
	 * Later, they will be deleted from the GC pool, for good.
	 * DynamicRouteEntries might get "resuscitated" from the pool.
	 * <p>
	 * Maps Integer() wrapped destintion addresses to RouteEntries.
	 * XXX this is actually another RouteTable instance...
	 */
	private final Hashtable gcPool;
	
	/**
	 * Gateways: hardwired connections to other networks.
	 * This map contains *only* the hardwired gateways.
	 * Nodes which are one hop away which are learned at run-time
	 * are <emph>not</emph> included in this map.  XXX I think
	 * that is a bug.
	 * <p>
	 * Maps Integer() wrapped destintion addresses to DynamicRouteEntries.
	 * XXX this is actually another RouteTable instance...
	 * <p>
	 * Only really used in the sendRouteMessages()
	 * function.  
	 *
	 * @see #sendRouteMessages(int)
	 */
	private final Hashtable gateways;
	
	public DynamicRouteTable()
	{
		this.gcPool = new Hashtable();
		this.gateways = new Hashtable();
	}


	/**
	 * XXX one-time function.
	 */
	public void configure(int nodeAddr, boolean doUpdates, String configFile)
		throws IOException, AddressFormatException, NoSuchNodeAddress
	{
		assert(doUpdates);
		assert(nodeAddr != 0);
		assert(configFile);
		
		RouteTableReader.read(nodeAddr, this, configFile);
	}

	/**
	 * Retrive the current route to get to the given active address.
	 */
	public RouteEntry getRoute(int activeDest)
	{
		return super.getRoute(activeDest);
	}

	public RouteEntry getRoute(Integer activeDest)
	{
		return super.getRoute(activeDest);
	}

	PACKAGE Hashtable gcPool()
	{
		return gcPool;
	}

	public Enumeration getNeighbors()
	{
		return gateways.elements();
	}

	public Enumeration getGateways()
	{
		return gateways.elements();
	}

	public int neighborCount()
	{
		return gateways.size();
	}

	/**
	 * Remove a route from the table.
	 */
	public void removeRoute(RouteEntry r)
	{
		super.removeRoute(r);
	}

	/**
	 * Add a route to the table.
	 * <p>
	 * XXX this is only invoked by RouteTableReader.  Don't
	 * call internally...
	 */
	public void addRoute(int activeDest, int activeNext, String physNext)
		throws NoSuchNodeAddress, AddressFormatException
	{
		// A dynamic table starts only with gateway systems
		if (activeDest == activeNext)
		{
			DynamicRouteEntry re = new DynamicRouteEntry(activeDest, activeNext, physNext);
			re.setCost(1);
			gateways.put(new Integer(activeDest), re);

			// of course, we know how to reach this guy...
			super.addRoute(re);
		}
	}

}
