/* 
 *  $Id: MD5.Java,v 1.2 2000/02/09 18:54:20 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_UTIL_PKG;

/**
 * Implementation of RSA's MD5 hash generator
 * <p>
 * This code is loosely based on the RSA reference implementation
 * and an early Java port of it by Santeri Paavolainen.
 * <p>
 * Objects of this class are single use only -- clients must
 * ensure that only one hash is computed, and sequentially too.
 * Hashes are returned as an array of 16 bytes.
 * <p>
 * Usage:
 * <pre>
 *   byte[] hash = MD5.hash(bytes);   // single pass mode
 * </pre>
 * or
 * <pre>
 *   MD5 digest = new MD5();          // multiple pass mode
 *   digest.update(bytes);
 *   digest.update(morebytes);
 *   byte[] hash = digest.hash();
 * </pre>
 * <p>
 * (Pat) Optimizations: combine all of the arrays into a single array
 * and use fixed offsets.
 * <p>
 *
 * @author David Wetherall
 */
public final class MD5 {
	/** 128-bit state */
	private final int state[] = new int[4];
  
  /** 64-byte transform pass */
	private final int tx[] = new int[16];

	/** 64-bit character count */
	private final int count[] = new int[2];

	/** 64-byte overflow buffer */
	private final byte buffer[] = new byte[64];

	private final static byte padding[] = {
		(byte) 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};
  
	public MD5() {
		state[0] = 0x67452301;
		state[1] = 0xefcdab89;
		state[2] = 0x98badcfe;
		state[3] = 0x10325476;

		count[0] = count[1] = 0;
	}

	private static int rotleft(int x, int n) {
		return (x << n) | (x >>> (32 - n));
	}

	private static int FF (int a, int b, int c, int d, int x, int s, int ac) {
		a +=  ((b & c) | (~b & d)) + x + ac;
		return rotleft(a, s) + b;
	}

	private static int GG (int a, int b, int c, int d, int x, int s, int ac) {
		a += ((b & d) | (c & ~d)) + x + ac;
		return rotleft(a, s) + b;
	}

	private static int HH (int a, int b, int c, int d, int x, int s, int ac) { 
		a += (b ^ c ^ d) + x + ac;
		return rotleft(a, s) + b;
	}

	private static int II (int a, int b, int c, int d, int x, int s, int ac) {  
		a += (c ^ (b | ~d)) + x + ac;
		return rotleft(a, s) + b;
	}

	private void transform (byte buf[], int shift) {
		int	a, b, c, d;

		a = state[0];
		b = state[1];
		c = state[2];
		d = state[3];

		/* swallow next 64 bytes as 16 ints */
		for (int i = 0, j = shift; i < tx.length; i++, j += 4) {
			tx[i] = (buf[j] & 0xff) | ((buf[j + 1] & 0xff) << 8) |
				((buf[j + 2] & 0xff) << 16) | ((buf[j + 3] & 0xff) << 24);
		}
    
		/* Round 1 */
		a = FF (a, b, c, d, tx[ 0],   7, 0xd76aa478); /* 1 */
		d = FF (d, a, b, c, tx[ 1],  12, 0xe8c7b756); /* 2 */
		c = FF (c, d, a, b, tx[ 2],  17, 0x242070db); /* 3 */
		b = FF (b, c, d, a, tx[ 3],  22, 0xc1bdceee); /* 4 */
		a = FF (a, b, c, d, tx[ 4],   7, 0xf57c0faf); /* 5 */
		d = FF (d, a, b, c, tx[ 5],  12, 0x4787c62a); /* 6 */
		c = FF (c, d, a, b, tx[ 6],  17, 0xa8304613); /* 7 */
		b = FF (b, c, d, a, tx[ 7],  22, 0xfd469501); /* 8 */
		a = FF (a, b, c, d, tx[ 8],   7, 0x698098d8); /* 9 */
		d = FF (d, a, b, c, tx[ 9],  12, 0x8b44f7af); /* 10 */
		c = FF (c, d, a, b, tx[10],  17, 0xffff5bb1); /* 11 */
		b = FF (b, c, d, a, tx[11],  22, 0x895cd7be); /* 12 */
		a = FF (a, b, c, d, tx[12],   7, 0x6b901122); /* 13 */
		d = FF (d, a, b, c, tx[13],  12, 0xfd987193); /* 14 */
		c = FF (c, d, a, b, tx[14],  17, 0xa679438e); /* 15 */
		b = FF (b, c, d, a, tx[15],  22, 0x49b40821); /* 16 */

		/* Round 2 */
		a = GG (a, b, c, d, tx[ 1],   5, 0xf61e2562); /* 17 */
		d = GG (d, a, b, c, tx[ 6],   9, 0xc040b340); /* 18 */
		c = GG (c, d, a, b, tx[11],  14, 0x265e5a51); /* 19 */
		b = GG (b, c, d, a, tx[ 0],  20, 0xe9b6c7aa); /* 20 */
		a = GG (a, b, c, d, tx[ 5],   5, 0xd62f105d); /* 21 */
		d = GG (d, a, b, c, tx[10],   9,  0x2441453); /* 22 */
		c = GG (c, d, a, b, tx[15],  14, 0xd8a1e681); /* 23 */
		b = GG (b, c, d, a, tx[ 4],  20, 0xe7d3fbc8); /* 24 */
		a = GG (a, b, c, d, tx[ 9],   5, 0x21e1cde6); /* 25 */
		d = GG (d, a, b, c, tx[14],   9, 0xc33707d6); /* 26 */
		c = GG (c, d, a, b, tx[ 3],  14, 0xf4d50d87); /* 27 */
		b = GG (b, c, d, a, tx[ 8],  20, 0x455a14ed); /* 28 */
		a = GG (a, b, c, d, tx[13],   5, 0xa9e3e905); /* 29 */
		d = GG (d, a, b, c, tx[ 2],   9, 0xfcefa3f8); /* 30 */
		c = GG (c, d, a, b, tx[ 7],  14, 0x676f02d9); /* 31 */
		b = GG (b, c, d, a, tx[12],  20, 0x8d2a4c8a); /* 32 */

		/* Round 3 */
		a = HH (a, b, c, d, tx[ 5],   4, 0xfffa3942); /* 33 */
		d = HH (d, a, b, c, tx[ 8],  11, 0x8771f681); /* 34 */
		c = HH (c, d, a, b, tx[11],  16, 0x6d9d6122); /* 35 */
		b = HH (b, c, d, a, tx[14],  23, 0xfde5380c); /* 36 */
		a = HH (a, b, c, d, tx[ 1],   4, 0xa4beea44); /* 37 */
		d = HH (d, a, b, c, tx[ 4],  11, 0x4bdecfa9); /* 38 */
		c = HH (c, d, a, b, tx[ 7],  16, 0xf6bb4b60); /* 39 */
		b = HH (b, c, d, a, tx[10],  23, 0xbebfbc70); /* 40 */
		a = HH (a, b, c, d, tx[13],   4, 0x289b7ec6); /* 41 */
		d = HH (d, a, b, c, tx[ 0],  11, 0xeaa127fa); /* 42 */
		c = HH (c, d, a, b, tx[ 3],  16, 0xd4ef3085); /* 43 */
		b = HH (b, c, d, a, tx[ 6],  23,  0x4881d05); /* 44 */
		a = HH (a, b, c, d, tx[ 9],   4, 0xd9d4d039); /* 45 */
		d = HH (d, a, b, c, tx[12],  11, 0xe6db99e5); /* 46 */
		c = HH (c, d, a, b, tx[15],  16, 0x1fa27cf8); /* 47 */
		b = HH (b, c, d, a, tx[ 2],  23, 0xc4ac5665); /* 48 */

		/* Round 4 */
		a = II (a, b, c, d, tx[ 0],   6, 0xf4292244); /* 49 */
		d = II (d, a, b, c, tx[ 7],  10, 0x432aff97); /* 50 */
		c = II (c, d, a, b, tx[14],  15, 0xab9423a7); /* 51 */
		b = II (b, c, d, a, tx[ 5],  21, 0xfc93a039); /* 52 */
		a = II (a, b, c, d, tx[12],   6, 0x655b59c3); /* 53 */
		d = II (d, a, b, c, tx[ 3],  10, 0x8f0ccc92); /* 54 */
		c = II (c, d, a, b, tx[10],  15, 0xffeff47d); /* 55 */
		b = II (b, c, d, a, tx[ 1],  21, 0x85845dd1); /* 56 */
		a = II (a, b, c, d, tx[ 8],   6, 0x6fa87e4f); /* 57 */
		d = II (d, a, b, c, tx[15],  10, 0xfe2ce6e0); /* 58 */
		c = II (c, d, a, b, tx[ 6],  15, 0xa3014314); /* 59 */
		b = II (b, c, d, a, tx[13],  21, 0x4e0811a1); /* 60 */
		a = II (a, b, c, d, tx[ 4],   6, 0xf7537e82); /* 61 */
		d = II (d, a, b, c, tx[11],  10, 0xbd3af235); /* 62 */
		c = II (c, d, a, b, tx[ 2],  15, 0x2ad7d2bb); /* 63 */
		b = II (b, c, d, a, tx[ 9],  21, 0xeb86d391); /* 64 */

		state[0] += a;
		state[1] += b;
		state[2] += c;
		state[3] += d;
	}

	public void update (byte buf[], int offset, int length) {
		int	index, partlen, i;

		/* compute number of bytes mod 64 */
		index = (count[0] >>> 3) & 0x3f;

		if ((count[0] += (length << 3)) < (length << 3)) count[1]++;
		count[1] += length >>> 29;

		partlen = 64 - index;

		if (length >= partlen) {
			System.arraycopy(buf, offset, buffer, index, partlen );

			transform(buffer, 0);

			for (i = partlen; (i + 63) < length; i+= 64)
				transform(buf, i);

			index = 0;
		} else 
			i = 0;

		/* buffer remaining input */
		if (i < length) {
			System.arraycopy(buf, i + offset, buffer, index, length - i);
		}
	}

	public void update (byte buf[]) {
		update(buf, 0, buf.length);
	}

	public void update (String s) {
		update(s.getBytes());
	}

	private static byte[] ints2bytes (int input[]) {
		byte out[] = new byte[input.length * 4];

		for (int i = 0, j = 0; i  < input.length; i++, j += 4) {
			out[j] = (byte) (input[i] & 0xff);
			out[j + 1] = (byte) ((input[i] >>> 8) & 0xff);
			out[j + 2] = (byte) ((input[i] >>> 16) & 0xff);
			out[j + 3] = (byte) ((input[i] >>> 24) & 0xff);
		}

		return out;
	}

	public byte[] hash() {
		byte bits[] = ints2bytes(count);
		int index = (int) ((count[0] >>> 3) & 0x3f);
		int padlen = (index < 56) ? (56 - index) : (120 - index);
		update(padding, 0, padlen);
		update(bits);	    
		byte hash[] = ints2bytes(state);

		// XXX unlink to prevent reuse
		// (Don't bother.  MD5 object gets dropped anyway.
		// state = null;
		// count = null;
		// tx = null;
		// buffer = null;
		return hash;
	}   

	public static byte[] hash(byte[] buf, int off, int len) {
		MD5 finger = new MD5();
		finger.update(buf, off, len);
		return finger.hash();
	}

	public static byte[] hash(byte[] buf) {
		return hash(buf, 0, buf.length);
	}

	public static String toString(byte hash[])
	{
		StringBuffer buf = new StringBuffer(16 * 2);
    
		for (int i = 0; i < 16; i++) {
			if ((hash[i] & 0xff) < 0x10)
				buf.append("0");
			buf.append(Integer.toString(hash[i] & 0xff, 16));
		}
    
		return buf.toString();
	}

	public static String toShortString(byte hash[]) 
	{
		StringBuffer buf = new StringBuffer((3 * 2) + 2);
    
		for (int i = 0; i < 3; i++)
		{
			if ((hash[i] & 0xff) < 0x10)
				buf.append("0");
			buf.append(Integer.toString(hash[i] & 0xff, 16));
		}
    
		buf.append("..");
		return buf.toString();
	}
}
