/*
 * Address.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 * NOTE: The toString() and fromString() functions in this class are
 * covered by: 
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 *
 */
package JANOS_NODEOS_PKG;

import java.net.InetAddress;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * An address object hides the details and implementation of
 * addressing in Janos.  Addresses can only be created through an
 * AddressFactory.  Only one instance of AddressFactory is created, to
 * control who is allowed to create Addresses.
 * <p>
 * An Address has two parts a logical "active network" address and
 * an underlying physical address.  The physical address in this
 * implementation of Address is an InetAddress and a UDP port number.
 * <p>
 * This implementation uses an IP host:port pair to indentify the
 * underlying address.  The Active address is a 32-bit IP-like number.
 * <p>
 * Address objects are immutable.  They must be immutable because
 * anyone(?) can read them out of routing tables..
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public class Address
{
	/**
	 * Flag to indicate if the factory has been given out or not.
	 * Once the factory has been created, no more are created.
	 */
	private static boolean gaveFactory = false;

	private final InetAddress iNetHostAddr; // physical
	private final int iNetPort; // physical

	private final int anAddress; // logical

	/**
	 * First caller wins.  The first caller of this method gets a
	 * handle on the AddressFactory.  Subsequent callers get
	 * diddly.  Weak-ass access control, thank you very much.
	 * <p>
	 * XXX implications of the 'static'-ness of this method?
	 */
	public static synchronized AddressFactory getAddressFactory()
	{
		if (gaveFactory)
			return null;

		gaveFactory = true;

		return new AddressFactory();
	}

	/**
	 * Construct a new Address.  The address has two parts an
	 * "active address" and an "internet address".  The "active address"
	 * is a 32-bit integer.  The "internet address" consists of an
	 * java.net.InetAddress and an iNetPort number.
	 * <p>
	 * XXX replace the InetAddress with a string that will be decoded
	 * by the NodeOS..
	 */
	PACKAGE Address(int anAddress, InetAddress iNetHostAddr, int iNetPort)
	{
		this.anAddress = anAddress;
		this.iNetHostAddr = iNetHostAddr;
		this.iNetPort = iNetPort;
	}

	PACKAGE Address(Address other)
	{
		this.anAddress = other.anAddress;
		this.iNetHostAddr = other.iNetHostAddr;
		this.iNetPort = other.iNetPort;
	}

	public int getANAddress()
	{
		return anAddress;
	}
	
	PACKAGE InetAddress iNetAddr() 
	{
		return iNetHostAddr;
	}

	PACKAGE int iNetPort()
	{
		return iNetPort;
	}

	public String toString()
	{
		return "Address[anAddr=" +toString(anAddress)+
			"; IP: " +iNetHostAddr.toString()+
			":" +iNetPort+ "]";
	}

	/**
	 * Generate a string representation of the given address.
	 * Generates standard dotted-quad notation.
	 */
	public static String toString (int a)
	{
		String dotted = Integer.toString(a & 0xff);
		dotted = Integer.toString((a >> 8) & 0xff) + "." + dotted;
		dotted = Integer.toString((a >> 16) & 0xff) + "." + dotted;
		dotted = Integer.toString((a >> 24) & 0xff) + "." + dotted;
		return dotted;
	}

	/**
	 * Generate a 32-bit integer from the given string address.
	 * The string must be a dotted-quad.
	 *
	 * @exception AddressFormatException if there aren't 4 parts, any of the 4 parts don't parse,
	 * or are out of range.
	 */
	public static int fromString(String name)
		throws AddressFormatException
	{
		StringTokenizer st = new StringTokenizer(name, ".");
		if (st.countTokens() != 4)
			throw new AddressFormatException("can't parse " +name);
		
		int[] b = new int[4];
		try
		{
			for (int i = 0; i < 4; i++) {
				b[i] = Integer.parseInt(st.nextToken());
				if ((b[i] < 0) || (b[i] > 255))
					throw new AddressFormatException("address part at "
									 +i+ " is out of range in " +name);
			}
		}
		catch (NumberFormatException e) 
		{
			throw new AddressFormatException("non-numeric address: " +name);
		}
		catch (NoSuchElementException _)
		{
			// This should never happen
			panic("trouble in Address.fromString!");
		}
		
		return ((b[0]<<24) | (b[1]<<16) | (b[2]<<8) | b[3]); 
	}

}
