/*
 * BufferHandle.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * A BufferHandle is a process's handle to a buffer.  A BufferHandle
 * object is seen by only one process, so that it is not shared
 * between processes, while the underlying Buffer object may be
 * shared between processes.  However, a Buffer object is never
 * directly accessible to a process; the Buffer can only be accessed
 * through the BufferHandle.
 * <p>
 * A process is charged for the memory occupied by all buffers reachable
 * from the process's BufferHandle objects.  To avoid paying for a buffer
 * which is no longer needed, a process can explicitly recycle a
 * BufferHandle object, which tells Janos that the process no longer
 * needs the buffer data associated with the BufferHandle.
 * Janos is then free to reuse the BufferHandle object for new buffer
 * data.  An access to a BufferHandle after it has been recycled behaves
 * unpredictably: it may continue to access the old buffer, it
 * may access another one of the processes's buffers, or it may
 * raise a NullPointerException.  However, such an access will never
 * read another process's buffer.
 * <p>
 * Currently, all access methods in BufferHandle are synchronized.  This
 * is inefficient.  The virtual machine can eliminate these locks in a
 * section of code by incrementing the buffer reference count before the
 * code section and decrementing the reference count after the code
 * section.  Unfortunately, this optimization cannot be expressed at
 * the Java source level.  (XXX is this something for Jason?)
 * <p>
 * Since a BufferHandle is visible only to one process, and a process
 * can only run on one processor at a time, simple preemption points
 * can eliminate the per-access synchronization.
 * <p>
 * Buffers are mutable.
 * <p>
 * BufferHandles are organized as circular doubly-linked lists.  At
 * any given time, each BufferHandle object belongs to exactly one
 * such list (all operations preserve this invariant).
 * The constructor creates a BufferHandle
 * which consists of a one element list.
 * <p>
 * A list of BufferHandles can be "sent" on an OutChannel.
 * An InChannel could return a list, too.  (But that doesn't happen).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public final class BufferHandle
{
	/**
	 * state.lock is used to protect the
	 * per-flow lists of buffer handles.
	 * 
	 */
	private final FlowState state;
	private Buffer buffer;
	private int offset;
	private int length;
	private BufferHandle next;
	private BufferHandle prev;


	/**
	 * Construct a new (empty) BufferHandle.
	 */
	public BufferHandle(Flow flow)
	{
		this.state = flow.state;
		next = this;
		prev = this;
	}

	/**
	 * Construct a new BufferHandle that points
	 * to the same data as the given handle.
	 * (Shallow copy of other bufferhandle).
	 */
	public BufferHandle(Flow flow, BufferHandle bh)
	{
		this(flow);
		this.copyFrom(bh); // does *not* do a deep copy.
	}

	/**
	 * Construct a new BufferHandle that points
	 * to the same data as the given handle.
	 * (Shallow copy of other bufferhandle).
	 */
	public BufferHandle(Flow flow, BufferHandle bh, int fromOffset, int fromLength)
	{
		this(flow);
		this.copyFrom(bh, fromOffset, fromLength); // does *not* do a deep copy.
	}

	/**
	 * Release the buffer held by this BufferHandle.
	 */
	public void reset()
	{
		setBuffer(null, 0, 0);
	}

	/**
	 * Point this BufferHandle to a new buffer which holds
	 * the specified array.  The old buffer is released.
	 *
	 * @allocation: Allocates a new Buffer
	 */
	public void newBuffer(int length)
	{
		setBuffer(new Buffer(new byte[length]), 0, length);
	}

	/**
	 * Set this BufferHandle's buffer to point to src's buffer.  The
	 * positions of the BufferHandles in their lists are not
	 * affected.
	 */
	public void copyFrom(BufferHandle src, int fromOffset, int fromLength)
		throws ArrayIndexOutOfBoundsException,
		       NegativeArraySizeException
	{
		// Two locks are required, but to avoid deadlock the first is
		// released before the second is acquired:
		Buffer srcBuffer;
		int srcOffset;
		int srcLength;
		synchronized(src)
		{
			srcBuffer = src.buffer;
			srcBuffer.acquire();
			srcOffset = src.offset;
			srcLength = src.length;
		}
		if(fromLength < 0)
			throw new NegativeArraySizeException();
		if(fromOffset < 0 ||
		   fromOffset + fromLength > srcLength ||
		   fromOffset + fromLength < 0)
			throw new ArrayIndexOutOfBoundsException();
		setBuffer(srcBuffer, srcOffset + fromOffset, fromLength);
	}

	/**
	 * Set this BufferHandle's buffer to point to src's buffer.  The
	 * positions of the BufferHandles in their lists are not
	 * affected.
	 */
	public void copyFrom(BufferHandle src)
	{
		// Two locks are required, but to avoid deadlock the first is
		// released before the second is acquired:
		Buffer srcBuffer;
		int srcOffset;
		int srcLength;
		synchronized(src)
		{
			srcBuffer = src.buffer;
			srcBuffer.acquire();
			srcOffset = src.offset;
			srcLength = src.length;
		}
		setBuffer(srcBuffer, srcOffset, srcLength);
	}

	public BufferHandle getNext()
	{
		return next;
	}

	public BufferHandle getPrev()
	{
		return prev;
	}

	/**
	 * Removes the object from the list that it is in.  This splits
	 * a list of length n into two lists, one of 1 (which contains
	 * this object), and one of length n-1 (which contains all the
	 * other objects).  If n==1, then removeFromList does nothing.
	 */
	public void removeFromList()
	{
		synchronized(state.lock)
		{
			BufferHandle nx = next;
			BufferHandle pv = prev;
			pv.next = nx;
			nx.prev = pv;
			next = this;
			prev = this;
		}
	}

	/**
	 * Removes the object from its current list and adds it to
	 * a new list.  The object is added after "list".
	 */
	public void addToList(BufferHandle list)
	{
		if(list == null) throw new NullPointerException();

		synchronized(state.lock)
		{

			// remove from old list:
			BufferHandle nx0 = next;
			BufferHandle pv0 = prev;
			pv0.next = nx0;
			nx0.prev = pv0;

			// add to new list:
			BufferHandle nx1 = list.next;
			list.next = this;
			nx1.prev = this;
			next = nx1;
			prev = list;
		}
	}

	// XXX why not length()?
	public synchronized int getLength()
	{
		return length;
	}

	public synchronized byte getByte(int index)
			throws ArrayIndexOutOfBoundsException
	{
		if(index < 0 || index >= length)
			throw new ArrayIndexOutOfBoundsException();
		return buffer.data[offset + index];
	}

	public synchronized void setByte(int index, byte val)
			throws ArrayIndexOutOfBoundsException
	{
		if(index < 0 || index >= length)
			throw new ArrayIndexOutOfBoundsException();
		buffer.data[offset + index] = val;
	}

	public synchronized int getInt(int index)
			throws ArrayIndexOutOfBoundsException
	{
		if(index < 0 || index + 4 > length)
			throw new ArrayIndexOutOfBoundsException();
		index += offset;
		byte[] buf = buffer.data;
		return (((buf[index++] & 0xFF) << 24) |
			((buf[index++] & 0xFF) << 16) |
			((buf[index++] & 0xFF) <<  8) |
			((buf[index]   & 0xFF)));
	}

	public synchronized void setInt(int index, int val)
			throws ArrayIndexOutOfBoundsException
	{
		if(index < 0 || index + 4 > length)
			throw new ArrayIndexOutOfBoundsException();
		index += offset;
		byte[] buf = buffer.data;
		buffer.data[index++] = (byte) (val >> 24);
		buffer.data[index++] = (byte) (val >> 16);
		buffer.data[index++] = (byte) (val >> 8);
		buffer.data[index] = (byte) (val);
	}

	public synchronized void getBytes(int srcPosition, byte[] dest,
			int destPosition, int destLength)
			throws ArrayIndexOutOfBoundsException
	{
		if(srcPosition < 0 || srcPosition + destLength > length)
			throw new ArrayIndexOutOfBoundsException();
		System.arraycopy(buffer.data, offset + srcPosition,
			dest, destPosition, destLength);
	}

	public synchronized void setBytes(int destPosition, byte[] src,
			int srcPosition, int srcLength)
			throws ArrayIndexOutOfBoundsException
	{
		if(destPosition < 0 || destPosition + srcLength > length)
			throw new ArrayIndexOutOfBoundsException();
		System.arraycopy(src, srcPosition,
			buffer.data, offset + destPosition, srcLength);
	}

	/**
	 * Get the Buffer associated with this BufferHandle.
	 */
	PACKAGE synchronized Buffer getBuffer(OffsetLength ol)
	{
		buffer.acquire();
		ol.offset = offset;
		ol.length = length;
		return buffer;
	}

	PACKAGE static final int aggregateLength(BufferHandle bh)
	{
		BufferHandle curr = bh;
		int retval = 0;

		do
		{
			retval += curr.getLength();
			curr = curr.getNext();
		} while(curr != bh);
		return retval;
	}
	
	PACKAGE static int aggregate(byte dest[], BufferHandle bh)
	{
		BufferHandle curr = bh;
		int pos = 0;

		do
		{
			int currLen = curr.getLength();

			curr.getBytes(0, dest, pos, currLen);
			pos += currLen;
			curr = curr.getNext();
		} while(curr != bh);
		return pos;
	}

	// The Buffer b must already be acquired before calling setBuffer.
	PACKAGE synchronized void setBuffer(Buffer b, int offset, int length)
	{
		// assert(b.acquired());
		if(buffer != null)
			buffer.release();
		this.buffer = b;
		this.offset = offset;
		this.length = length;
	}

	protected void finalize() throws Throwable
	{
		if(buffer != null)
			buffer.release();
	}

	public String toString()
	{
		return "BufferHandle[state="
			+ state
			+ "; buffer="
			+ buffer
			+ "; offset="
			+ offset
			+ "; length="
			+ length
			+ "]";
	}
	
#ifdef DEBUG
	public final void warnRefCount(int r, String s)
	{
		if(buffer.getRefCount() > r)
			TRACE(Trace.NODEOS_BUFH, "WARNING: ref ct=" + buffer.getRefCount() + " (" + s + ")");
	}

	public final void assertRefCountIs(int r)
	{
		assert(buffer.getRefCount() == r);
	}
#endif // DEBUG

}
