/*
 * CommSpaceHandle.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.lang.ArrayIndexOutOfBoundsException;

/**
 * Each flow has exactly one CommSpaceHandle.
 * The CommSpaceHandle points to exactly one CommSpace.
 * <p>
 * XXX Semantics during a failed take() are probably wrong.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public final class CommSpaceHandle
{
	private CommSpace commSpace;

	public static CommSpaceHandle newCommSpace(int size)
	{
		return new CommSpaceHandle(new CommSpace(size));
	}

	public static CommSpaceHandle newCommSpace(CommSpaceHandle csh, int size)
	{
		assert(csh); // XXX
		assert(csh.empty()); // XXX

		csh.handleFor(new CommSpace(size));
		return csh;
	}

	public CommSpaceHandle()
	{
		commSpace = null;
	}

	/**
	 * Copy constructor.  Only used when passing a
	 * csh to a child flow.
	 */
	PACKAGE CommSpaceHandle(CommSpaceHandle other)
	{
		assert(other);
		// this.owner = owner
		commSpace = other.commSpace;
	}

	PACKAGE CommSpaceHandle(CommSpace commSpace)
	{
		// this.owner = owner;
		this.commSpace = commSpace;
	}

	PACKAGE boolean empty()
	{
		return commSpace == null;
	}

	PACKAGE void handleFor(CommSpace commSpace)
	{
		this.commSpace = commSpace;
	}

	public int size()
	{
		return commSpace.size();
	}

	public int resize(int newsize)
	{
		NOTIMPLEMENTED("resize()");

		// XXX only doable by allocator of underlying CommSpace object

		// watch out for shrinkage.
	}

	/**
	 * Write a CommSpaceElement into the CommSpace.
	 * Overwrites any pre-existing element at that index.
	 */
	public void write(int index, CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		// assert(elem.owner == Flow.currentFlow());
		commSpace.write(index, elem);
	}
	
	/**
	 * XXX is it necessary to copy the CommSpaceElement out
	 * of the CommSpace??
	 */
	private static CommSpaceElement copy(CommSpaceElement e)
	{
		return (e == null)?(null):(new CommSpaceElement(e));
	}

	public CommSpaceElement take(int index)
		throws ArrayIndexOutOfBoundsException
	{
		return copy(commSpace.take(index));
	}

	public CommSpaceElement waitToTake(int index)
		throws ArrayIndexOutOfBoundsException,
		InterruptedException
	{
		return copy(commSpace.waitToTake(index));
	}
	
	public CommSpaceElement read(int index)
		throws ArrayIndexOutOfBoundsException
	{
		return copy(commSpace.read(index));
	}

	public CommSpaceElement swap(int index, CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		return copy(commSpace.swap(index, elem));
	}
}
