/*
 * FlowClassLoader.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Vector;
import java.util.zip.ZipFile; // XXX
import java.util.zip.ZipEntry; // XXX
import java.io.File; // XXX
import java.io.FileInputStream; // XXX
import java.io.BufferedInputStream; // XXX
import java.io.IOException; // XXX

/**
 * Loads class files for a Flow.  This ClassLoader is critical to the security 
 * of the system.  It must prevent dynamic code from installing
 * untrusted code in trusted packages or overriding trusted code.
 * <p>
 * There are three kinds of classes that a Flow is concerned with:
 * system classes, local classes and dynamic classes.  All are
 * optionally subject to package access control checks.
 * <p>
 * System classes are classes that get loaded once, and get loaded by
 * the JVM's root class loader.  The "system" classes are defined on a
 * package basis.  See Main for the code that initializes the set of
 * system packages (this include the base Java packages: java.lang,
 * java.util; and the various Janos NodeOS classes: janos.nodeos,
 * janos.util, etc.)  The set of system packages is recorded in the
 * static variable "systemPackages".
 * <p>
 * Local classes are classes that get loaded off of the node's local
 * disk, but get loaded into each Flow.  These consist of Applications
 * loaded at end nodes.  And maybe some day extensions will fall in
 * this category(?).  Like system classes, local classes are
 * categorized on a package basis.  The set of local packages is
 * implicitly those that are neither dynamic nor system packages.
 * XXX: It may be advantageous to put other packages (e.g, the ANTS
 * utils package) in this category to make all the static variables
 * per-flow.
 * <p>
 * Dynamic classes are classes that are installed via a CodeGroup.
 * They exclusively apply to the unit of code that is shipped around.
 * <p>
 * Aside from these classifications, access to packages can be
 * controlled.  By default flow access control is disabled.  Once
 * enabled, then only system classes and local classes listed in the
 * package access table will be made visible to the flow.  This
 * prevents Flows from accessing system classes they shouldn't have
 * access to.
 * <p>
 * XXX Not sure if this is correct wrt the new JDK1.2 ClassLoader
 * semantics (e.g., should we be overriding findClass??).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE final class FlowClassLoader 
	extends ClassLoader
{
	/**
	 * A mapping from classname to bytecode.  This is the set of
	 * classes that are loaded over the net into this ClassLoader.
	 */
	private final Hashtable dynamicClasses = new Hashtable();

	/**
	 * Table of package names.  Packages included in this table
	 * are loaded by the System ClassLoader.
	 * <p>
	 * XXX this is static so that (1) it is shared by all
	 * Flows in the system and (2) so that it can be initialized
	 * by Main.
	 *
	 * @see #loadClass(String,boolean)
	 * @see #addSystemPackage(String)
	 */
	private static final Hashtable systemPackages = new Hashtable();

	/**
	 * A table of Java packages which this flow has "access" to.
	 * Default to just the janos.nodeos package.
	 * <p>
	 * Due to the design of ClassLoaders, this table is only
	 * checked for class references from dynamically loaded
	 * classes.
	 */
	private final Hashtable okPackages = new Hashtable();
	private boolean packageControl = false;

	private static final Vector pathcomps = new Vector(8);
	
#ifdef DEBUG
	/**
	 * For coherent debugging, need to know what flow I'm
	 * associated with.
	 */
	private final Flow flow;
#endif

	private static class ClassLoaderInfo
		extends InfoHook
	{
		/**
		 * This is a pointer to FlowClassLoader.pathcomps.
		 * It is *only* here so that Janos will compile with
		 * the broken JDK1.2 javac.  Directly accessing the
		 * enclosing classes private static  pathcomps field
		 * causes the compiler to think the private final flow
		 * field is never initialized.
		 */
		private final Vector pathcomps;

		public ClassLoaderInfo(Vector pathcomps)
		{
			super("ClassLoader");
			this.pathcomps = pathcomps;
		}

		public String[] getInfo()
		{
			String retval[];

			synchronized(pathcomps)
			{
				retval = new String[pathcomps.size()];
				for( int i = 0; i < pathcomps.size(); i++ )
				{
					retval[i] = pathcomps.elementAt(i).
						toString();
				}
			}
			return retval;
		}
	}

	static {
		Status.addInfoHook(new ClassLoaderInfo(pathcomps));
	}
	
	/**
	 * Initialize the pathcomps vector to the elements of
	 * the global system classpath.
	 */
	static { // XXX this is bogus, should be configured somewhere
		String cp = java.lang.System.getProperty("java.class.path");
		
		int i;
		while ((i = cp.indexOf(File.pathSeparatorChar)) != -1) {
			if (i == 0)
				// XXX right thing to do?
				pathcomps.addElement(".");
			else
				pathcomps.addElement(cp.substring(0, i));
			cp = cp.substring(i+1);
		}
		pathcomps.addElement(cp);
	}

	/**
	 * Make a deep copy of the ClassSpec and push into
	 * the table of name to bytecode mappings.
	 * <p>
	 * FlowClassLoaders are only created when a new Flow() is
	 * created.  The ClassSpec elements have to be duplicated
	 * because the parameter is from the parent process's
	 * memory.
	 *
	 * @param classSpecs array of name and bytecode pairs.
	 */
	PACKAGE FlowClassLoader(Flow flow, ClassSpec[] classSpecs)
	{
		DEBUGONLY(this.flow = flow);
		
		for(int i = 0; i < classSpecs.length; i++)
		{
			// XXX duplicate string?  Shared between processes?
			String s = new String(classSpecs[i].name);
			byte[] b = new byte[classSpecs[i].bytecode.length];
			System.arraycopy(classSpecs[i].bytecode, 0, b, 0,
				b.length);
			dynamicClasses.put(s, b);
		}
	}

	/**
	 * Add a new class to an existing flow.  This should
	 * be called from inside the flow that the class is
	 * being added to.
	 * <p>
	 * Neither argument may be null.
	 *
	 * @param name the FQN of the class to be added
	 * @param bytecode the bytecode representation of that class
	 */
	PACKAGE void addClass(String name, byte[] bytecode)
	{
		TRACE(Trace.NODEOS_CLASSLOAD, "Adding class " +name+ " to flow " + flow);
		assert(name != null);
		assert(bytecode != null);
		dynamicClasses.put(name, bytecode);
	}

	/**
	 * Allow access to a Java package from this ClassLoader.
	 */
	public void allowPackage(String pkgPrefix)
	{
		TRACE(Trace.NODEOS_CLASSLOAD, "Allowing package " +pkgPrefix+ " in flow " + flow);
		assert(pkgPrefix);
		okPackages.put(pkgPrefix, pkgPrefix);
	}

	/**
	 * Turn package access controls on or off.
	 */
	public void enablePackageControls(boolean doit)
	{
		if (doit)
		{
			TRACE(Trace.NODEOS_CLASSLOAD, "Enabled package controls in flow " + flow);
		}
		else
		{
			TRACE(Trace.NODEOS_CLASSLOAD, "Disabled package controls in flow " + flow);
		}
		
		packageControl = doit;
	}

	public static void addSystemPackage(String pkgName)
	{
		TRACE(Trace.NODEOS_CLASSLOAD, "Adding system package " +pkgName+ ".");
		// Map name -> name (just need something non-null in the map).
		systemPackages.put(pkgName, pkgName);
	}

	public static boolean isSystemPackage(String pkgName)
	{
		return (systemPackages.get(pkgName) != null);
	}

	/**
	 * Read the bytecode for the given class off the local file system.
	 * Uses the pathcomps static variable as the "classpath" to
	 * search.
	 * <p>
	 * Bogus algorithm.  Should know if entry is zipfile or not, and
	 * re-use that information (instead of rediscovering it).
	 */
	private static byte[] readLocalBytecode(String classname)
		throws ClassNotFoundException
	{
		assert(classname);

		// XXX no cache, as all the caching is done elsewhere...
		// check in cache first
		// byte [] p = (byte [])classcache.get(classname);
		//
		// if (p != null)
		//	return p;
		
		String cname = classname.replace('.', '/').concat(".class");
		
		// iterate through all components and look for .class file
		for (int i = 0; i < pathcomps.size(); i++)
		{
			BufferedInputStream file = null;
			int size = 0;
			boolean foundit = false;
			String compname = (String)pathcomps.elementAt(i);

			foundit = false;

			try
			{
				// a zip file ?
				ZipFile  zip   = new ZipFile(compname);
				ZipEntry entry = zip.getEntry(cname);
				
				if (entry != null)
				{
					file = new BufferedInputStream(zip.getInputStream(entry));
					size = (int)entry.getSize();
					assert(size != -1); // XXX handle more gracefully
					foundit = true;
				}
			}
			catch (java.io.IOException ioe)
			{
				// ah, well
			}
		
			if (!foundit)
			{
				try
				{
					String fn  = compname.concat("/").concat(cname);
					File f = new File(fn);
					file = new BufferedInputStream(new FileInputStream(f));
					size = (int)f.length(); 
					foundit = true;
				}
				catch (java.io.FileNotFoundException fnfe)
				{
					// ah, well
				}
			}

			// Try the next classpath entry
			if (!foundit)
				continue;
			
			assert(file != null);

			// Try to read the file
			byte[] result = new byte[size];
			try
			{
				if (file.read(result) != result.length)
					throw new ClassNotFoundException("File " +file+ " is too short for " +classname); // XXX better exception
			}
			catch (java.io.IOException ioe)
			{
				throw new ClassNotFoundException("IO exception loading "
								 +file+ " for " +classname+
								 ": " +ioe.getMessage());
			}
			
			return result;
		}
		
		throw new ClassNotFoundException("Could not find bytecode for " +classname+ ".");
	}

	/**
	 * Look for the given class on the local filesystem.
	 * <p>
	 * XXX this should be native and should take advantage of
	 * the search code builtin to the JVM.
	 */
	private Class readLocalClass(String name)
		throws ClassNotFoundException
	{
		byte[] bytes;

		TRACE(Trace.NODEOS_CLASSLOAD, "Reading class " +name+ " from disk in flow " + flow);

		bytes = readLocalBytecode(name);
		assert(bytes != null);
		return defineClass(name, bytes, 0, bytes.length);
	}

	/**
	 * @see #loadClass(String,boolean)
	 */
	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		return this.loadClass(name, false);
	}

	/**
	 * The standard ClassLoader loadClass interface.
	 * <p>
	 * This loadClass looks for already-loaded classes, looks in the
	 * local dynamicClasses, and then checks the package name, then
	 * looks in the system class loader.
	 * <p>
	 * XXX Currently any code defined in the dynamicClasses will be
	 * loaded, *regardless of the package it is in*.  This is
	 * because a Flow really has no way of knowing what package
	 * a legit protocol was defined in.  No easy solution.  It is a
	 * bad problem because dynamic protocols can define code to load
	 * into packages like "janos.nodeos"!
	 * <p>
	 * XXX also, there is a problem that defineClass() is called before
	 * the package access check is done.  Thus <clinit> might run before
	 * the access control check is made.  That would be bad.
	 */
	public synchronized Class loadClass(String name, boolean resolve)
		throws ClassNotFoundException
	{
		Class cl;

		TRACE(Trace.NODEOS_CLASSLOAD, "Loading class " +name+ " in flow " + flow);

		// If its already loaded, just get it and get out.
		cl = findLoadedClass(name);
		if (cl != null)
		{
			TRACE(Trace.NODEOS_CLASSLOAD, " ... class " +name+ " was already loaded.");
			// Fall through to resolve and return
		}
		else
		{
			int offset;
			String pkgName;
			
			assert(cl == null);

			// Get the package name
			offset = name.lastIndexOf('.');
			if (offset == 0)
			{
				// XXX No-name package is always inaccessible?
				throw new ClassNotFoundException("The no-name package is always inaccessible.");
			}
			pkgName = name.substring(0, offset);
						
			if (isSystemPackage(pkgName))
			{
				/*
				 * Access to system packages can be controlled.
				 */
				if ((packageControl)
				    && (okPackages.get(pkgName) == null))
				{
					throw new ClassNotFoundException("Access to system package "
									 +pkgName+ " is not enabled.");
				}

				/*
				 * System packages are loaded by the system
				 * class loader.
				 */
				cl = findSystemClass(name);
				assert(cl != null); // throws otherwise

				TRACE(Trace.NODEOS_CLASSLOAD, " ... class " +name+ " is a system class.");

				// Fall through to resolve and return
			}
			else
			{
				// See if such a class was dynamically installed
				byte[] b = (byte[]) dynamicClasses.get(name);
				if(b != null)
				{
					// XXX package control doesn't work here...

					// XXX May throw weird exceptions/errors, no?
					cl = defineClass(name, b, 0, b.length);
					assert(cl != null);
					TRACE(Trace.NODEOS_CLASSLOAD, " ... class " +name+ " loaded from dynamicClasses.");
				
					// XXX Can drop bytecode from dynamicClasses now...

					// fall through to resolve and return
				}
				else
				{
					assert(cl == null);

					/*
					 * Check that the package is accessible.
					 */
					if (packageControl
					    && (okPackages.get(pkgName) == null))
						throw new ClassNotFoundException("Access to the " +pkgName+ " package is not enabled.");

					cl = readLocalClass(name);
					assert(cl != null); // throws otherwise

					TRACE(Trace.NODEOS_CLASSLOAD, " ... class " +name+ " loaded from disk.");
				}
				
				assert(cl != null);
			}

			assert(cl != null);
		}
		
		assert(cl != null);
		
		if (resolve)
			resolveClass(cl);

		return cl;
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
	}

#ifdef DEBUG
	/**
	 * When debugging is enabled, the toString()
	 * method goes to some lengths to avoid printing
	 * the object default identifier (i.e, @82e9350)
	 * because that screws up the regression tests.
	 */
	private static int dbg_currentIdNumber = 0;
	private int dbg_idNumber = dbg_currentIdNumber++;

	public String toString()
	{
		return FlowClassLoader.class.getName() + "#" + dbg_idNumber;
	}
#else
	public String toString()
	{
		return "FlowClassLoader@0x" + Integer.toHexString(this.hashCode());
	}
#endif
}
