/*
 * InChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

import java.io.IOException; // XXX
import java.net.DatagramSocket; // XXX
import java.net.DatagramPacket; // XXX

#ifdef BENCH
import JANOS_UTIL_PKG.CycleTimer;
import JANOS_UTIL_PKG.Semaphore;
#endif

//#define CHECK_DUPLICATE_CLASSIFIERS

/**
 * Java wrapper for an InChannel.
 * <p>
 * Implemented on top of UDP for the moment.
 *
 * @see CutThroughChannel
 * @see OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public final class InChannel
{
	/**
	 * Death Packet. XXX.
	 */
	public static final byte[] exitPacket = {-0x22, -0x53, -0x42, -0x11};

	/**
	 * This is a kernel-global variable.  It maps Demultiplex
	 * keys to InChannels.
	 *
	 * XXX global variable
	 */
	private static final Vector classifierTable = new Vector();

	/**
	 * The flow this receive channel is associated with.
	 */
	PACKAGE final Flow flow;

	/**
	 * The demultiplex key associated with this channel.
	 */
	private final DemultiplexKey classifier;

	/**
	 * The ThreadPool to send incoming packets to.
	 * ThreadPool.addWork() had best be protected kernel code
	 * (i.e., final).
	 */
	private final ThreadPool threadPool;

	/**
	 * The callback for delivery.
	 */
	private final DeliverFunc deliverFunc;

	public InChannel(Flow flow,
			 DemultiplexKey classifier,
			 ThreadPool threadPool,
			 DeliverFunc deliverFunc)
	{
		this.flow = flow;
		this.classifier = classifier;
		this.threadPool = threadPool;
		this.deliverFunc = deliverFunc;

		assert(classifier);
		// XXX doesn't sufficiently protect the kernel...
		if((flow == null)
		   || (threadPool == null)
		   || (classifier == null))
			throw new NullPointerException();

		addClassifier(this);

		flow.addChannel(this);
	}
	
	private static synchronized void addClassifier(InChannel newChan)
	{
		// Insert in sorted order...
		int i;
		for (i = 0; i < classifierTable.size(); i++)
		{
			InChannel entry = (InChannel) classifierTable.elementAt(i);
#ifdef CHECK_DUPLICATE_CLASSIFIERS
			if (entry.classifier.equals(newChan.classifier))
				throw new IllegalStateException(
					"Cannot add duplicate classifiers");
#endif
			if (entry.classifier.length() <= newChan.classifier.length())
				break;
		}
		TRACE(Trace.NODEOS_DISPATCH, "Adding packet classifier " +newChan.classifier);
		classifierTable.insertElementAt(newChan, i);
	}

	private static synchronized void remClassifier(InChannel deadChan)
	{
		int i;

		for( i = 0; i < classifierTable.size(); i++ )
		{
			InChannel entry =
				(InChannel)classifierTable.elementAt(i);

			if( entry.classifier.equals(deadChan.classifier) )
			{
				classifierTable.removeElementAt(i);
				break;
			}
		}
	}

	/**
	 * An InfoHook class to report the set of packet classifiers
	 */
	private static class PacketClassifierInfo
		extends InfoHook
	{
		private final Vector classifierTable;
		
		private PacketClassifierInfo(Vector classifierTable)
		{
			super("PacketClassifier");
			this.classifierTable = classifierTable;
		}

		public String[] getInfo()
		{
#if 0
			String[] result;
			StringBuffer buf = new StringBuffer(200);
			
			synchronized(classifierTable)
			{
				result = new String[classifierTable.size()];
				for(int i = 0; i < classifierTable.size(); i++)
				{
					InChannel elem =
						(InChannel)classifierTable.
						elementAt(i);
					
					buf.setLength(0);
					buf.append(elem.classifier.toString());
					buf.append(" -> ");
					buf.append(elem.deliverFunc.
						   toString());
					result[i] = buf.toString();
				}
			}
			return result;
#endif
			return null;
		}
	}

	static {
		Status.addInfoHook(new PacketClassifierInfo(classifierTable));
	}
	
	/**
	 * XXX
	 * @see #msgLoop(int)
	 */
	public static void msgLoop(Address localAddr)
		throws IOException
	{
		assert(localAddr);
		assert(localAddr.iNetPort() != 0);
		msgLoop(localAddr.iNetPort());
	}

	/**
	 *  This must be run before any Channel objects are created.
	 *  <p>
	 *  XXX: shouldn't really be public
         *
	 *  @allocations
         */
	public static void msgLoop(int recvPort)
		throws IOException
	{
		// Set up a "packet of death":
		DemultiplexKey exitKey = new DemultiplexKey(1024) {
			public int matches(byte[] buf, int packetOffset, int packetLength) {
						   return compareSlice(buf, packetOffset, packetLength,
								       0, exitPacket);
					   }
		};
		
		DatagramSocket socket = new DatagramSocket(recvPort);

		// We need to have at least one packet around to
		// receive data before the data gets demultiplexed:
		UdpBuffer.addBuffers(1);

		for(;;)
		{
			boolean sentIt;
			UdpBuffer b = UdpBuffer.getFree();
			assert(b != null);
			
			BENCHONLY(processTimer.end());

			b.receive(socket);

			BENCHONLY(sendToRecvTimer.end());
			BENCHONLY(recvToSendTimer.start());
			BENCHONLY(recvToRecvTimer.end());
			BENCHONLY(recvToRecvTimer.start());
			BENCHONLY(processTimer.start());

			TRACE(Trace.NODEOS_INCOMING, "in!");
			TRACE(Trace.NODEOS_INCOMING, "received: " + Debug.dbgPacketSummary(b));
			sentIt = demultiplex(b);

			if(!sentIt
			   && (exitKey.matches(b.data, 0, b.length()) != 0))
			{
				b.release();
				break;
			}
			else if(!sentIt)
			{
				Thread.yield();
				TRACE(Trace.NODEOS_INCOMING, "Can't schedule pkt: port=" +recvPort+ "; data=" +Debug.dbgPacketSummary(b));
				TRACE(Trace.NODEOS_BUFH, "dropped refcnt=" +b.getRefCount());
			}
			b.release();
		}

		TRACE(Trace.NODEOS_INCOMING, "recv port " +recvPort+ ": exiting");
	}

	/**
	 * I sure hope this gets better
	 * <p>
	 * XXX I broke this.  The return value is true if the
	 * packet was dispatched *and* the target had enough
	 * resources to handle the packet.  It returns false
	 * if the packet wasn't matched *or* it was matched
	 * but the target didn't have enough resources.
	 */
	PACKAGE static boolean demultiplex(UdpBuffer b)
	{
		/*
		 * Assumes entries are sorted by length (longest first).
		 */
		for(int i = 0; i < classifierTable.size(); i++)
		{
			int payloadStart;
			InChannel entry = (InChannel)
				classifierTable.elementAt(i);
			assert(entry.classifier); // constructor must verify this
			payloadStart = entry.classifier.matches(b.data,
								0,
								b.length());
			if (payloadStart != 0)
			{
				TRACE(Trace.NODEOS_INCOMING, "demultiplexed to: " +entry.classifier);
				return entry.scheduleDelivery(b, 0, b.length(),
							      payloadStart);
			}
		}
		
		return false;
	}

	/**
	 * Schedule the given buffer b for delivery to this InChannel.
	 * A BufferHandle wrapping the packet is set up (from packetStart
	 * to packetLength).  The payloadStart parameter is just a *hint*
	 * passed to the delivery function.  To be meaningful, it should
	 * be relative to the packetStart.
	 * <p>
	 * Called by demultiplex (above) and by InChannelHandle.push().
	 *
	 * @return true if the buffer was scheduled, false if the
	 * target had insuffient resources.
	 */
	PACKAGE boolean scheduleDelivery(Buffer b, int packetStart, int packetLength,
					 int payloadStart)
	{
		// XXX I think I broke something.. 
		if(!b.charge(this.flow.state))
		{
			// flow is out of buffer space:
			TRACE(Trace.NODEOS_INCOMING, "out of buffer space");
			return false;
		}

		PacketWorker worker = PacketWorker.getFree(flow);
		assert(worker != null); // XXX 

		worker.deliverFunc = this.deliverFunc;
		worker.bufferHandle.setBuffer(b, packetStart, packetLength);

		worker.payloadStart = payloadStart;
		BENCHONLY(threadSchedTimer.start());
		threadPool.addWork(worker.workItem);
		return true;
	}

	public String toString()
	{
		return "InChannel[flow="
			+ flow
			+ "; classifier="
			+ classifier
			+ "; threadPool="
			+ threadPool
			+ "; deliverFunc="
			+ deliverFunc
			+ "]";
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		remClassifier(this);
	}
	
	protected void finalize()
	{
		remClassifier(this);
	}

#ifdef BENCH
	public static CycleTimer processTimer = new CycleTimer("InChannel processing");
	public static CycleTimer sendToRecvTimer = new CycleTimer("InChannel send-to-recv");
	public static CycleTimer recvToSendTimer = new CycleTimer("InChannel recv-to-send");
	public static CycleTimer recvToRecvTimer = new CycleTimer("InChannel recv-to-recv");
	public static CycleTimer threadSchedTimer = new CycleTimer("InChannel thread schedule");
	public static Semaphore cheatLock = new Semaphore();
	public static byte[] cheatBuffer = new byte[65536];
	public static int cheatLen;
#endif //BENCH
}


