/*
 * OutChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.DatagramSocket;
import java.net.DatagramPacket;

#ifdef BENCH
import JANOS_UTIL_PKG.CycleTimer;
#endif

/**
 * Java wrapper for a Channel.
 * <p>
 * Implemented on top of UDP for the moment.
 *
 * @see CutThroughChannel
 * @see InChannel
 */
public final class OutChannel
{
	PACKAGE final Flow flow;

	/**
	 */
	private final DatagramPacket sendPacket;

	/**
	 */
	private final byte[] gatherBuffer;

	/**
	 */
	private final OffsetLength offsetLength;

	/**
	 * Not final because we null it out on destroy...
	 */
	private /*final*/ DatagramSocket sendSocket;

#ifdef BENCH
	public static CycleTimer udpSendTimer = new CycleTimer("OutChannel udp send");
#endif //BENCH

	public OutChannel(Flow flow,
			  Address sendAddr)
		throws IOException
	{
		if(flow == null || sendAddr == null)
			throw new NullPointerException();

		this.flow = flow;

		gatherBuffer = new byte[UdpBuffer.MAX_PACKET_LEN];
		offsetLength = new OffsetLength();
		sendPacket = new DatagramPacket(gatherBuffer, 0,
						sendAddr.iNetAddr(), sendAddr.iNetPort());
		sendSocket = new DatagramSocket();

		flow.addChannel(this);
	}

	public synchronized void send(BufferHandle packet)
		throws IOException
	{
		final Buffer buf = packet.getBuffer(offsetLength);

		// Copy the field to avoid race conditions in destroy()
		final DatagramSocket sendSocket = this.sendSocket;

		// sendSocket is only null if the OutChan has been destroyed.
		if (sendSocket == null)
		{
			buf.release();
			return;
		}

		if(packet.getNext() == packet && offsetLength.offset == 0)
		{
			// Only one element in the aggregate:
			sendPacket.setData(buf.data);
			sendPacket.setLength(offsetLength.length);
		}
		else
		{
			synchronized(flow.state.lock)
			{
				int len = BufferHandle.aggregate(gatherBuffer,
								 packet);

				sendPacket.setData(gatherBuffer);
				sendPacket.setLength(len);
			}
		}
		TRACE(Trace.NODEOS_OUTGOING, "sending: dest_port = " +sendPacket.getPort()+ " data = " +Debug.dbgPacketSummary(sendPacket.getData(), 0, sendPacket.getLength()));

		BENCHONLY(InChannel.recvToSendTimer.end());
		BENCHONLY(InChannel.sendToRecvTimer.start());
		BENCHONLY(udpSendTimer.start());
		/* 
		 * We might get a SocketException here which causes an
		 * explosion further up the call chain 
		 */
		try
		{
#ifndef BENCH
			sendSocket.send(sendPacket);
#else //BENCH
			InChannel.cheatLen = sendPacket.getLength();
			System.arraycopy(sendPacket.getData(), 0, InChannel.cheatBuffer, 0, InChannel.cheatLen);
			InChannel.cheatLock.signal();
#endif //BENCH
		}
		catch(Exception e)
		{
			TRACE(Trace.NODEOS_OUTGOING, "Exception when trying to send on socket");
		}
		BENCHONLY(udpSendTimer.end());

		buf.release();
	}

	PACKAGE void destroy()
	{
		DatagramSocket x;

		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		// To avoid race conditions between null/close
		x = sendSocket;
		sendSocket = null;
		if( x != null )
			x.close();
	}
}


