/*
 * Thread.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * A Janos Thread. <code>java.lang.Thread</code> is not visible to
 * threads running on the NodeOS.
 * <p>
 * A Janos thread is never explicitly created by a Protocol, only
 * implicitly by ThreadGroups.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public class Thread
{
	/**
	 * The java.lang.Thread this Janos thread is implemented
	 * on top of.
	 * <p>
	 * This is just a cheap hack until we get the native support
	 * working correctly.
	 */
	private NativeThread nThread;

	/**
	 * Per-thread state.
	 * <p>
	 * This is part of the NodeOS spec.  Probably doesn't need to
	 * be part of the Java-level interface (users of this should
	 * just subclass Janos Thread).
	 */
	private Object obj;
	
	/**
	 * The ThreadPool this thread is a part of.
	 */
	private final ThreadPool tPool;

	/**
	 * For creating an implicit thread.
	 * <p>
	 * All thread constructors are package private.
	 */
	PACKAGE Thread(ThreadPool tPool)
	{
		super();
		assert(tPool != null);
		this.obj = null;
		this.tPool = tPool;
		this.nThread = new NativeThread(this);
		this.nThread.start();
	}

	/**
	 * For creating an explicit thread.
	 * <p>
	 * All thread constructors are package private.
	 */
	PACKAGE Thread(ThreadPool tPool, Runnable r)
	{
		super();
		this.obj = null;
		this.tPool = tPool;
		this.nThread = new NativeThread(this, r);
		this.nThread.start();
	}

	/**
	 * Set the name of the thread
	 */
	public void setName(String name)
	{
		this.nThread.setName(name);
	}

	/**
	 * Get the name of the thread
	 */
	public String getName()
	{
		return this.nThread.getName();
	}

	/**
	 * Package-private run method that all threads start in
	 * (once kicked).
	 */
	PACKAGE void run()
	{
		tPool.doWork();
	}

	/**
	 * Return the Thread object associated with the currently
	 * executing thread of control.
	 *
	 * @see currentThread()
	 * @return a Thread object associated with the current thread.
	 */
	public static Thread self() 
	{
		return currentThread();
	}

	/**
	 *
	 * @see #self()
	 */
	public static Thread currentThread()
	{
		java.lang.Thread jth;
		NativeThread nth;

		jth = NativeThread.currentThread();
		assert(jth instanceof NativeThread);
		nth = (NativeThread)jth;
		
		return nth.janosThread;
	}

	public Flow getFlow()
	{
		return this.tPool.getFlow();
	}

	/**
	 * The mighty directed yield.
	 * <p>
	 * The scheduler may honor the argument.  Or it may
	 * laugh at your suggestion and put the target thread
	 * in a black hole just to spite you.
	 *
	 * @param yieldTo the thread that should be run in deference to the current thread.
	 */
	public static void yieldTo(Thread yieldTo)
	{
		NativeThread.yieldTo((yieldTo == null)?(null):(yieldTo.nThread));
	}
	
	/**
	 * An undirected yield.
	 * <p>
	 * Equivalent to <code>yield(null)</code>.
	 */
	public static void yield()
	{
		yieldTo(null);
	}

	/**
	 * Put the current thread to sleep for at least
	 * the given amount of time.
	 */
	public static void sleep(long millis, int nanos)
		throws InterruptedException
	{
		NativeThread.sleep(millis, nanos);
	}

	public String toString()
	{
		return "Janos NativeThread@0x" + Integer.toHexString(this.hashCode());
	}

	public void setObject(Object obj)
	{
		this.obj = obj;
	}

	public Object getObject()
	{
		return this.obj;
	}

	public static void interrupt(Thread target)
	{
		panic("HEY!  DON'T YOU INTERRUPT ME!");
	}
}
