/*
 * UdpBuffer.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;

/**
 * A UDP implementation of Buffer.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
PACKAGE class UdpBuffer
	extends Buffer
{
	PACKAGE static UdpBuffer freeList;
	PACKAGE static final int MAX_PACKET_LEN = 65536;

	private static int usedBuffers = 0;
	private static int maxUsedBuffers = 0;
	private static int absMaxUsedBuffers = 0;
	private static int availableBuffers = 0;
	private static int totalBuffers = 0;

	private static final float maxUseDecay = 1.25f;
	private static final float decayOffset = 2.0f;
	
	private final DatagramPacket udpPacket;
	PACKAGE UdpBuffer nextFree;
	private FlowState state;
	private int len;

	private static class UdpBufferInfo
		extends InfoHook
	{
		private UdpBufferInfo()
		{
			super("UdpBuffer");
		}
		
		public String[] getInfo()
		{
			String retval[] = new String[5];

			retval[0] = "Current used buffers: "
				+ UdpBuffer.usedBuffers();
			retval[1] = "Current max buffers: "
				+ UdpBuffer.maxUsedBuffers();
			retval[2] = "Absolute max buffers: "
				+ UdpBuffer.absMaxUsedBuffers();
			retval[3] = "AvailableBuffers: "
				+ UdpBuffer.availableBuffers();
			retval[4] = "TotalBuffers: "
				+ UdpBuffer.totalBuffers();
			return retval;
		}
	}
	
	static {
		Status.addInfoHook(new UdpBufferInfo());
	}
	
	/**
	 * Add a bunch of buffers to the free buffer queue.
	 */
	PACKAGE static synchronized void addBuffers(int numBuffers)
	{
		totalBuffers += numBuffers;
	}

	/**
	 * Remove a bunch of buffers from the free buffer queue
	 */
	PACKAGE static synchronized void remBuffers(int numBuffers)
	{
		totalBuffers -= numBuffers;
	}
	
	/**
	 * Get the next available UdpBuffer.
	 * <p>
	 * Returns null if no buffers are available.
	 *
	 * @allocation Might have to allocate a buffer if none are readily
	 * available
	 */
	PACKAGE static synchronized UdpBuffer getFree()
	{
		UdpBuffer b = freeList;

		if( (b == null) &&
		    (maxUsedBuffers < totalBuffers) )
		{
			/*
			 * No free buffers on the queue, but we haven't used
			 * up the total available buffers
			 */
			b = new UdpBuffer();
			usedBuffers++; // XXX addToFreeList will decrement it
			addToFreeList(b);
			availableBuffers++;
		}
		if( b != null )
		{
			freeList = b.nextFree;
			b.nextFree = null;
			assert(b.state == null);
			
			usedBuffers++;
			maxUsedBuffers = Math.max(usedBuffers, maxUsedBuffers);
			absMaxUsedBuffers = Math.max(absMaxUsedBuffers,
						     maxUsedBuffers);
		}
		return b;
	}
	
	/**
	 * UdpBuffers can only be allocated by addBuffers()
	 * <p>
	 * Each UdpBuffer is MAX_PACKET_LEN bytes big.
	 *
	 * @see #addBuffers()
	 */
	private UdpBuffer()
	{
		super(new byte[MAX_PACKET_LEN]);
		
		udpPacket = new DatagramPacket(data, data.length);
	}
	
	/**
	 * Charge the given flow for this buffer.  This
	 * buffer will be associated with the flow (via
	 * this.state).
	 * <p>
	 * Returns true if the given flow can pay for this buffer.
	 */
	PACKAGE boolean charge(FlowState state)
	{
		synchronized(state.lock)
		{
			if (state.freeBuffers == 0)
				return false;

			/*
			 * If the buffer was charged to one Flow and is now
			 * being charged to another, we have to credit the
			 * first one.
			 */
			if( this.state != null )
				credit(this.state);

			// Associate with flow so we can credit it
			this.state = state;

			// Charge the flow
			state.freeBuffers--;
			assert(this.nextFree == null);
			this.acquire();

			return true;
		}
	}

	/**
	 * Credit the flow associated with this buffer.
	 */
	private void credit(FlowState state)
	{
		assert(state);
		synchronized(state.lock)
		{
			state.freeBuffers++;
		}
	}
	
	/**
	 * Put this buffer back in the recycle list (we're
	 * done with it).
	 */
	PACKAGE void recycle()
	{
		FlowState flowState = this.state;

		acquire();
		addToFreeList(this);
		if( flowState != null )
		{
			/* Only credit if we're part of a state */
			credit(flowState);
		}
	}

	/**
	 * Add the given flow to this global free list.
	 */
	private static synchronized void addToFreeList(UdpBuffer b)
	{
		assert(b.nextFree == null);
		usedBuffers--;
		/* Check if we should put the buffer back on the free list */
		if( (((float)usedBuffers * maxUseDecay + decayOffset) >=
		     (float)(maxUsedBuffers - 1)) )
		{
			/*
			 * Our buffer use is still pretty close to the max
			 * so we put it back on the list for later use
			 */
			b.state = null;
			b.nextFree = freeList;
			freeList = b;
		}
		else
		{
			/*
			 * We're using fewer buffers so we can just dump this
			 * one and free up the memory
			 */
			maxUsedBuffers--;
			availableBuffers--;
		}
	}

	PACKAGE void receive(DatagramSocket socket)
		throws IOException
	{
#ifndef BENCH
		udpPacket.setLength(data.length);
		socket.receive(udpPacket);
		len = udpPacket.getLength();
#else //BENCH
		InChannel.cheatLock.block();
		len = InChannel.cheatLen;
		System.arraycopy(InChannel.cheatBuffer, 0, data, 0, len);
#endif //BENCH
	}

	PACKAGE int length()
	{
		return len;
	}

	PACKAGE static int usedBuffers()
	{
		return usedBuffers;
	}

	PACKAGE static int maxUsedBuffers()
	{
		return maxUsedBuffers;
	}

	PACKAGE static int absMaxUsedBuffers()
	{
		return absMaxUsedBuffers;
	}

	PACKAGE static int availableBuffers()
	{
		return availableBuffers;
	}

	PACKAGE static int totalBuffers()
	{
		return totalBuffers;
	}

}
