/* 
 * RouteTable.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ROUTE_PKG;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;

import JANOS_NODEOS_PKG.Status;
import JANOS_NODEOS_PKG.InfoHook;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Routing table.  Maps 32-bit integer "active addresses" to
 * RouteEntries.  A RouteEntry contains the core Address object
 * which, in turn, contains the "real" address of the target node.
 * RouteEntry can be subclassed to add more information to the table.
 * <p>
 * XXX most the hashtable used in this class is indexed by 'ints'.
 * But, that requires allocating (and dropping) a lot of "Integer"
 * objects.  Could be optimized by moving to a more pure hashtable (or
 * putting the Integer in the RouteTable.  Also, would benefit from
 * parameterized tables.
 * <p>
 * By design, this class defines a read-only RouteTable for users
 * in other packages.  Access to WriteableRouteTable is required
 * to make updates to the RouteTable.
 * <p>
 * Note that the read-only-ness of this class depends on the
 * immutablity of base RouteEntries.
 *
 * @see WriteableRouteTable
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu>Janos Ministry of Development</a>
 */
public class RouteTable 
{
	/**
	 * The table of all known routes.
	 */
	private final Hashtable routes;

	/**
	 * The table of immediate neighbors.
	 */
	private final Hashtable neighbors;

	/**
	 * InfoHook class that provides routing info via
	 * the NodeOS Status interface.
	 */
	private class RouteTableInfo
		extends InfoHook
	{
		private RouteTableInfo()
		{
			super("routes");  //  XXX paramterize name
		}

		public String[] getInfo()
		{
			String retval[];
			Enumeration e;
			int ct;

			// synchronization problem between creating
			// tables of size X and reading enumerations
			// from them...  use a Vector instead of String[].

			ct = 2 + routeCount() + neighborCount();
			retval = new String[ct];

			ct = 0;
			retval[ct++] = "Neighbors:";
			e = getNeighbors();
			for (int i = 0; i < neighborCount(); i++)
			{
				assert(e.hasMoreElements());
				retval[ct++] = e.nextElement().toString();
			}
			retval[ct++] = "Routes:";
			e = getRoutes();
			for (int i = 0; i < routeCount(); i++)
			{
			assert(e.hasMoreElements());
				retval[ct++] = e.nextElement().toString();
			}

			return retval;
		}
	}

	/**
	 * Construct a table for routes.  The initial table is empty.
	 * Routes can be added through <code>addRoute()</code> or
	 * ...
	 * <p>
	 * XXX take a size hint for the hashtable.
	 * 
	 */
	public RouteTable()
	{
		this.routes = new Hashtable();
		this.neighbors = new Hashtable();

		// XXX Should parameterize in case multiple route tables.
		Status.addInfoHook(new RouteTableInfo());
	}
	
	public Enumeration getRoutes()
	{
		return routes.elements();
	}

	public Enumeration getNeighbors()
	{
		return neighbors.elements();
	}

	public int routeCount()
	{
		return routes.size();
	}

	public int neighborCount()
	{
		return neighbors.size();
	}

	/**
	 * A convenience function for casting objects
	 * returned from the routes hashtable.
	 * <p>
	 * XXX Using parameterized Hashtables would eliminate this,
	 * methinks.
	 * XXX can't this assert() that all o's are instances
	 * of RouteEntry or null?
	 */
	private static RouteEntry castToRouteEntry(Object o)
	{
		if (o instanceof RouteEntry)
		{
			return (RouteEntry)o;
		}
		return null;
	}
	
	/**
	 * Retrive the current route to get to the given active address.
	 */
	public RouteEntry getRoute(int activeDest)
	{
		RouteEntry result;
		assert(activeDest != 0);
		result = getRoute(new Integer(activeDest));
		assert((result == null)
		       || (result.getLogicalDest() == activeDest));
		return result;
	}

	public RouteEntry getRoute(Integer activeDest)
	{
		RouteEntry result;
		assert(activeDest);
		result = castToRouteEntry(routes.get(activeDest));
		assert((result == null)
		       || (result.getLogicalDest() == activeDest.intValue()));
		TRACE(Trace.ROUTE_UPDATE, "getRoute(" +Address.toString(activeDest.intValue())+ ") -> " +result);
		return result;
	}

	/**
	 * Remove a route from the table.
	 * <p>
	 * External access through WriteableRouteTable
	 */
	PACKAGE void package_removeRoute(RouteEntry r)
	{
		assert(r);

		TRACE(Trace.ROUTE_UPDATE, "package_removeRoute(" +r+ ")");

		// XXX check that it exists?
		routes.remove(new Integer(r.getLogicalDest()));
	}

	/**
	 * Add a route to the table.
	 * <p>
	 * External access through WriteableRouteTable
	 */
	PACKAGE void package_addRoute(RouteEntry route) 
	{
		Object old;
		assert(route);

		// See if its a neighbor
		if (route.getLogicalDest() == route.getLogicalNext())
			neighbors.put(new Integer(route.getLogicalDest()), route);

		TRACE(Trace.ROUTE_UPDATE, "addRoute(" +route+ ")");

		old = routes.put(new Integer(route.getLogicalDest()), route);

		// XXX
		assert(old == null);
	}
}
