/*
 * DemultiplexKeyTester.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

package JANOS_TESTS_NODEOS_PKG;

import JANOS_TESTS_PKG.*;
import JANOS_NODEOS_PKG.*;

public class DemultiplexKeyTester extends Tester
{
	/**
	 * Subclass of DemultiplexKey to test
	 */
	private class TestKey extends DemultiplexKey
	{
		public TestKey(int length)
		{
			super(length);
		}

		/**
		 * Provides access to the Demultiplex.compareByte static method
		 */
		public final int compByte(byte[] packetBuf,
					  int po,
					  int pl,
					  int o,
					  byte v)
		{
#ifdef EXPOSED
			return DemultiplexKey.compareByte(packetBuf,
							  po,
							  pl,
							  o,
							  v);
#else
			return 0;
#endif
		}

		/**
		 * Provides access to the DemultiplexKey.compareSlice
		 * static method
		 */
		public final int compSlice(byte[] packetBuf,
					   int po,
					   int pl,
					   int o,
					   byte[] s)
		{
#ifdef EXPOSED
			return DemultiplexKey.compareSlice(packetBuf,
							   po,
							   pl,
							   o,
							   s);
#else
			return 0;
#endif
		}
		
		public int matches(byte[] data, int po, int pl)
		{
			return 0;
		}
	}

	/**
	 * Report the first occurrence of value in bytes
	 *
	 * @param bytes The array to scan through
	 * @param value The value to search for in bytes
	 * @return The index of the first occurrence of value, or -1 if its
	 * not found
	 */
	private int firstByte(byte bytes[], byte value)
	{
		int retval = -1;
		int lpc;

		for( lpc = 0; (lpc < bytes.length) && (retval == -1); lpc++ )
		{
			if( bytes[lpc] == value )
				retval = lpc;
		}
		return retval;
	}

	/**
	 * Generate a random array of bytes for a key
	 *
	 * @param length The length of the array
	 * @return An array of bytes with random element values
	 */
	private byte[] generateKeyBytes(int length)
	{
		byte retval[] = new byte[length];
		int lpc;

		for( lpc = 0; lpc < length; lpc++ )
		{
			retval[lpc] = (byte)(Math.random() * 10);
		}
		return retval;
	}

	/**
	 * Create an array from a section of another
	 *
	 * @param key The array to create the slice from
	 * @param offset The start of the section in key to copy
	 * @param length The length of the section in key to copy
	 * @return A new byte array that has the values from the section
	 * of key specified by the parameters
	 */
	private byte[] generateSlice(byte key[], int offset, int length)
	{
		byte retval[] = new byte[length];

		System.arraycopy(key, offset, retval, 0, length);
		return retval;
	}

	/**
	 * Create a byte array that won't match the specified subsection of key
	 *
	 * @param key The byte array to generate the bad slice from
	 * @param offset The start of the section in key to create the bad
	 * slice from
	 * @param length The length of the section in key
	 * @return A byte array that won't match the section in key specified
	 * by the parameters
	 */
	private byte[] generateBadSlice(byte key[], int offset, int length)
	{
		byte retval[] = new byte[length];
		int lpc;

		for( lpc = 0; lpc < length; lpc++ )
		{
			retval[lpc] = (byte)(key[offset + lpc] + 1);
		}
		return retval;
	}
	
	public void testUnit() throws TestFailedException
	{
		byte keyBytes[], keySlice[], badSlice[];
		int keyLength = 16;
		TestKey tk;
		int lpc;

		keyBytes = this.generateKeyBytes(keyLength);
		tk = new TestKey(keyLength);
		this.enterSection(ENV_EXPOSED, "Accessor");
#ifdef EXPOSED
		this.test(tk.length() == keyLength);
		
		this.enterSection(ENV_EXPOSED, "Comparisons");
		int testOffsets[] =
		{
			0,
			1,
			keyLength - 1
		};
		for( lpc = 0; lpc < testOffsets.length; lpc++ )
		{
			this.test(tk.compByte(keyBytes,
					      0,
					      keyLength,
					      testOffsets[lpc],
					      keyBytes[testOffsets[lpc]])
				  ==
				  (testOffsets[lpc] + 1));
			this.test(tk.compByte(keyBytes,
					      0,
					      keyLength,
					      testOffsets[lpc],
					      (byte)
					      (keyBytes[testOffsets[lpc]]
						    + 1))
				  ==
				  0);
		}

		keySlice = this.generateSlice(keyBytes,
					      keyLength / 2,
					      keyLength / 2);
		badSlice = this.generateBadSlice(keyBytes,
						 keyLength / 2,
						 keyLength / 2);
		this.test(tk.compSlice(keyBytes,
				       0,
				       keyLength,
				       keyLength / 2,
				       keySlice) ==
			  ((keyLength / 2) + keySlice.length));
#endif
	}
}
